import importlib
import json
import sys
import tempfile
import urllib
from pathlib import Path


def import_python(path, module_name=None):
    path = Path(path)
    return importlib.machinery.SourceFileLoader(
        module_name or path.stem, str(path)
    ).load_module()


def import_ipynb(path, module_name):
    """Import a Jupyter notebook as a module."""

    # Create a new module
    module = type(sys)(module_name)
    module.__file__ = str(path.resolve())

    # Add the module to sys.modules
    sys.modules[module_name] = module

    # Parse the notebook JSON
    with path.open("r", encoding="utf-8") as f:
        notebook = json.load(f)

    # Extract code from code cells
    code_cells = [
        cell for cell in notebook.get("cells", []) if cell.get("cell_type") == "code"
    ]

    # Combine all code cells, skipping cells marked with "# skip-import" comment
    code = ""
    for cell in code_cells:
        cell_source = "".join(cell.get("source", []))
        if "# skip-import" not in cell_source:
            code += cell_source + "\n\n"

    # Execute the code in the module's namespace
    try:
        exec(code, module.__dict__)
    except Exception as e:
        # Clean up if execution fails
        if module_name in sys.modules:
            del sys.modules[module_name]
        raise ImportError(f"Error executing notebook code: {e}") from e

    return module


def import_from_path(path, module_name="__main__"):
    """
    Import a Python script or Jupyter notebook as a module.

    Args:
        path (Path or str): pathlib.Path object or string pointing to a Python script (.py) or
                           Jupyter notebook (.ipynb)

    Returns:
        module: The imported module

    Raises:
        ImportError: If the file cannot be imported
        ValueError: If the file type is not supported
    """
    path = Path(path)

    # Check if path exists
    if not path.exists():
        raise ImportError(f"File not found: {path}")

    # Handle based on file extension
    if path.suffix == ".py":
        return import_python(path, module_name)
    elif path.suffix == ".ipynb":
        return import_ipynb(path, module_name)
    else:
        raise ValueError(
            f"Unsupported file type: {path.suffix}. Only .py and .ipynb files are supported."
        )


def import_from_gist(user, gist):
    path = f"https://gist.githubusercontent.com/{user}/{gist}/raw/"
    req = urllib.request.Request(path, method="GET")
    resp = urllib.request.urlopen(req)
    if resp.code != 200:
        raise ValueError(f"failed to fetch gist: HTTP {resp.code}")
    with tempfile.NamedTemporaryFile() as ntfh:
        path = Path(ntfh.name)
        path.write_text(resp.read().decode("ascii"))
        module = import_python(path)
        return module



import pytest

from xorq.backends.pyiceberg.tests.conftest import QUOTES_TABLE_NAME
from xorq.tests.util import assert_frame_equal


def test_create_table(iceberg_con, quotes_df):
    # the iceberg_con fixture creates the table
    assert QUOTES_TABLE_NAME in iceberg_con.list_tables()

    t = iceberg_con.table(QUOTES_TABLE_NAME)
    assert_frame_equal(t.execute(), quotes_df, check_dtype=False)


def test_create_and_drop_table(iceberg_con, trades_df):
    table_name = "trades"

    iceberg_con.create_table(table_name, trades_df, overwrite=True)
    assert table_name in iceberg_con.list_tables()

    iceberg_con.drop_table(table_name)
    assert table_name not in iceberg_con.list_tables()


def test_select(quotes_table):
    expected = quotes_table.execute()["symbol"].to_frame()
    actual = quotes_table.select("symbol").execute()

    assert_frame_equal(actual, expected, check_dtype=False)


def test_limit(quotes_table):
    actual = quotes_table.limit(5).execute()
    assert len(actual) == 5


def test_filter(quotes_table):
    df = quotes_table.execute()
    expected = df[df["symbol"] == "GOOGL"]
    actual = quotes_table.filter(quotes_table.symbol == "GOOGL").execute()

    assert_frame_equal(actual, expected, check_dtype=False)


def test_full_selection(quotes_table):
    actual = (
        quotes_table.select("symbol", "bid")
        .filter(quotes_table.symbol == "GOOGL")
        .limit(5)
        .execute()
    )
    assert len(actual) == 5
    assert actual.columns.tolist() == ["symbol", "bid"]
    assert actual["symbol"].eq("GOOGL").sum() == 5


def test_raises_not_implemented_error(quotes_table):
    with pytest.raises(NotImplementedError):
        quotes_table.count().execute()



import pyarrow as pa
import pytest

import xorq.api as xo
import xorq.ibis_yaml
import xorq.ibis_yaml.utils
from xorq.expr.udf import pyarrow_udwf
from xorq.vendor import ibis


def test_built_in_udf_properties(compiler):
    t = xo.table({"a": "int64"}, name="t")

    @xo.udf.scalar.builtin
    def add_one(x: int) -> int:
        return x + 1

    expr = t.mutate(new=add_one(t.a))
    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    original_mutation = expr.op()
    roundtrip_mutation = roundtrip_expr.op()

    original_udf = original_mutation.values["new"]
    roundtrip_udf = roundtrip_mutation.values["new"]

    assert original_udf.__func_name__ == roundtrip_udf.__func_name__
    assert original_udf.__input_type__ == roundtrip_udf.__input_type__
    assert original_udf.dtype == roundtrip_udf.dtype
    assert len(original_udf.args) == len(roundtrip_udf.args)

    for orig_arg, rt_arg in zip(original_udf.args, roundtrip_udf.args):
        assert orig_arg.dtype == rt_arg.dtype


def test_compiler_raises(compiler):
    t = xo.table({"a": "int64"}, name="t")

    @xo.udf.scalar.python
    def add_one(x: int) -> int:
        pass

    expr = t.mutate(new=add_one(t.a))
    with pytest.raises(NotImplementedError):
        compiler.to_yaml(expr)


@pytest.mark.xfail(
    reason="UDFs do not have the same memory address when pickled/unpickled"
)
def test_built_in_udf(compiler):
    # (Pdb) diffs[3][2].args[0] == diffs[3][1].args[0]
    # False
    # (Pdb) diffs[3][2].args[0]
    # <ibis.expr.operations.relations.Project object at 0x7ffff48f53d0>
    # (Pdb) diffs[3][2].args[0].args
    # (<ibis.expr.operations.relations.UnboundTable object at 0x7ffff48f5310>, {'a': <ibis.expr.operations.relations.Field object at 0x7ffff490cbb0>, 'new': <tests.test_udf.add_one_1 object at 0x7ffff48f5550>})
    # (Pdb) diffs[3][1].args[0].args
    # (<ibis.expr.operations.relations.UnboundTable object at 0x7ffff48f45f0>, {'a': <ibis.expr.operations.relations.Field object at 0x7ffff490cb40>, 'new': <tests.test_udf.add_one_1 object at 0x7ffff48f49b0>})
    t = xo.table({"a": "int64"}, name="t")

    @xo.udf.scalar.builtin
    def add_one(x: int) -> int:
        pass

    expr = t.mutate(new=add_one(t.a))
    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    print(f"Original {expr}")
    print(f"Roundtrip {roundtrip_expr}")
    xorq.ibis_yaml.utils.diff_ibis_exprs(expr, roundtrip_expr)

    assert roundtrip_expr.equals(expr)


def test_pandas_udf_properties(compiler):
    t = xo.table({"a": "int64", "b": "float64"}, name="t")

    @xo.udf.make_pandas_udf(
        schema=xo.schema({"a": int, "b": float}),
        return_type=xo.expr.datatypes.float64,
        name="multiply_add",
    )
    def multiply_add(df):
        return df["a"] * df["b"] + df["a"]

    expr = t.mutate(result=multiply_add.on_expr(t))
    yaml_dict = compiler.to_yaml(expr)

    roundtrip_expr = compiler.from_yaml(yaml_dict)

    original_mutation = expr.op()
    roundtrip_mutation = roundtrip_expr.op()
    original_udf = original_mutation.values["result"]
    roundtrip_udf = roundtrip_mutation.values["result"]

    assert original_udf.__func_name__ == roundtrip_udf.__func_name__
    assert original_udf.__input_type__ == roundtrip_udf.__input_type__
    assert original_udf.dtype == roundtrip_udf.dtype
    assert len(original_udf.args) == len(roundtrip_udf.args)


@pytest.fixture
def df():
    batch = pa.RecordBatch.from_arrays(
        [
            pa.array([0, 1, 2, 3, 4, 5, 6]),
            pa.array([7, 4, 3, 8, 9, 1, 6]),
            pa.array(["A", "A", "A", "A", "B", "B", "B"]),
        ],
        names=["a", "b", "c"],
    )

    return batch.to_pandas()


def test_udwf_roundtrip(compiler, df):
    @pyarrow_udwf(
        schema=ibis.schema({"a": float}),
        return_type=ibis.dtype(float),
        alpha=0.9,
    )
    def exp_smooth(self, values: list[pa.Array], num_rows: int) -> pa.Array:
        results = []
        curr_value = 0.0
        values = values[0]
        for idx in range(num_rows):
            if idx == 0:
                curr_value = values[idx].as_py()
            else:
                curr_value = values[idx].as_py() * self.alpha + curr_value * (
                    1.0 - self.alpha
                )
            results.append(curr_value)

        return pa.array(results)

    con = xo.connect()
    t = con.register(df, table_name="t")

    expr = t.select(
        t.a,
        udwf=exp_smooth.on_expr(t).over(ibis.window(group_by=ibis._.c)),
    ).order_by(t.a)

    yaml_dict = compiler.to_yaml(expr)

    profiles = {con._profile.hash_name: con}
    roundtrip_expr = compiler.from_yaml(yaml_dict, profiles=profiles)

    roundtrip_expr.execute()



import functools
import itertools
import operator
from collections import defaultdict
from typing import Any, Callable

import pyarrow as pa
import toolz
from opentelemetry import trace

from xorq.backends.xorq import connect as xo_connect
from xorq.common.utils.otel_utils import tracer
from xorq.common.utils.rbr_utils import (
    copy_rbr_batches,
    instrument_reader,
)
from xorq.vendor import ibis
from xorq.vendor.ibis import Expr, Schema
from xorq.vendor.ibis.common.collections import (
    FrozenDict,
    FrozenOrderedDict,
)
from xorq.vendor.ibis.common.graph import Graph
from xorq.vendor.ibis.expr import operations as ops
from xorq.vendor.ibis.expr.format import fmt, render_schema
from xorq.vendor.ibis.expr.operations import Node, Relation


def replace_cache_table(node, kwargs):
    if kwargs:
        node = node.__recreate__(kwargs)

    if isinstance(node, CachedNode):
        return node.parent.op().replace(replace_cache_table)
    elif isinstance(node, RemoteTable):
        return node.remote_expr.op().replace(replace_cache_table)
    else:
        return node


# https://stackoverflow.com/questions/6703594/is-the-result-of-itertools-tee-thread-safe-python
class SafeTee(object):
    """tee object wrapped to make it thread-safe"""

    def __init__(self, teeobj, lock):
        self.teeobj = teeobj
        self.lock = lock

    def __iter__(self):
        return self

    def __next__(self):
        with self.lock:
            return next(self.teeobj)

    def __copy__(self):
        return SafeTee(self.teeobj.__copy__(), self.lock)

    @classmethod
    def tee(cls, iterable, n=2):
        """tuple of n independent thread-safe iterators"""
        from itertools import tee
        from threading import Lock

        lock = Lock()
        return tuple(cls(teeobj, lock) for teeobj in tee(iterable, n))


def recursive_update(obj, replacements):
    if isinstance(obj, Node):
        if obj in replacements:
            return replacements[obj]
        else:
            return obj.__recreate__(
                {
                    name: recursive_update(arg, replacements)
                    for name, arg in zip(obj.argnames, obj.args)
                }
            )
    elif isinstance(obj, (tuple, list)):
        return tuple(recursive_update(o, replacements) for o in obj)
    elif isinstance(obj, dict):
        return {
            recursive_update(k, replacements): recursive_update(v, replacements)
            for k, v in obj.items()
        }
    else:
        return obj


def replace_source_factory(source: Any):
    def replace_source(node, _, **kwargs):
        if "source" in kwargs:
            kwargs["source"] = source
        return node.__recreate__(kwargs)

    return replace_source


def make_native_op(node):
    # FIXME: how to reference let.Backend.name?
    if node.source.name != "xorq":
        raise ValueError(f"Expected 'xorq' backend, but got {node.source.name!r}")
    sources = node.source._sources
    native_source = sources.get_backend(node)
    if native_source.name == "xorq":
        raise ValueError("Expected a native backend, but got 'let' backend")

    def replace_table(_node, _kwargs):
        return sources.get_table_or_op(
            _node, _node.__recreate__(_kwargs) if _kwargs else _node
        )

    return node.replace(replace_table).to_expr()


class Tag(ops.Relation):
    schema: Schema
    parent: Any
    metadata: FrozenOrderedDict = FrozenOrderedDict()
    values = FrozenDict()

    @property
    def tag(self):
        return self.metadata.get("tag")

    def __dask_tokenize__(self):
        from xorq.common.utils.dask_normalize.dask_normalize_expr import (
            normalize_seq_with_caller,
        )

        return normalize_seq_with_caller(
            self.schema,
            self.parent,
            self.metadata,
            caller="normalize_tag",
        )


class DatabaseTableView(ops.DatabaseTable):
    pass


class CachedNode(DatabaseTableView):
    parent: Any = None
    cache: Any = None


gen_name_namespace = "rbr-placeholder"
gen_name = toolz.compose(
    # some engines simply truncate long names
    operator.itemgetter(slice(0, 35)),
    functools.partial(ibis.util.gen_name, gen_name_namespace),
)


class RemoteTable(DatabaseTableView):
    remote_expr: Expr = None

    @classmethod
    def from_expr(cls, con, expr, name=None):
        name = name or gen_name()
        return cls(
            name=name,
            schema=expr.schema(),
            source=con,
            remote_expr=expr,
        )


def into_backend(expr, con, name=None):
    return RemoteTable.from_expr(con=con, expr=expr, name=name).to_expr()


class FlightExpr(DatabaseTableView):
    input_expr: Expr = None
    unbound_expr: Expr = None
    make_server: Callable = None
    make_connection: Callable = None
    do_instrument_reader: bool = False

    @classmethod
    def validate_schema(cls, input_expr, unbound_expr):
        from xorq.common.utils.graph_utils import walk_nodes

        (dt, *rest) = walk_nodes(ops.UnboundTable, unbound_expr)
        if rest or not isinstance(dt, ops.UnboundTable):
            raise ValueError("unbound_expr must contain exactly one UnboundTable")
        if dt.schema != input_expr.schema():
            raise ValueError(
                "Schema of unbound_expr does not match schema of input_expr"
            )

    @classmethod
    def from_exprs(
        cls,
        input_expr,
        unbound_expr,
        make_server=None,
        make_connection=None,
        name=None,
        **kwargs,
    ):
        from xorq.flight import FlightServer

        def roundtrip_cloudpickle(obj):
            import cloudpickle

            return cloudpickle.loads(cloudpickle.dumps(obj))

        cls.validate_schema(input_expr, unbound_expr)
        return cls(
            name=name or gen_name(),
            schema=unbound_expr.schema(),
            source=input_expr._find_backend(),
            input_expr=input_expr,
            unbound_expr=roundtrip_cloudpickle(unbound_expr),
            make_server=make_server or FlightServer,
            make_connection=make_connection or xo_connect,
            **kwargs,
        )

    def to_rbr(self, do_instrument_reader=None):
        from xorq.flight.action import AddExchangeAction
        from xorq.flight.exchanger import (
            UnboundExprExchanger,
        )

        if do_instrument_reader is None:
            do_instrument_reader = self.do_instrument_reader

        def inner(flight_exchange):
            rbr_in = flight_exchange.input_expr.to_pyarrow_batches()
            if do_instrument_reader:
                rbr_in = instrument_reader(rbr_in, "input: ")
            with flight_exchange.make_server() as server:
                client = server.client
                unbound_expr_exchanger = UnboundExprExchanger(
                    flight_exchange.unbound_expr
                )
                client.do_action(
                    AddExchangeAction.name,
                    unbound_expr_exchanger,
                    options=client._options,
                )
                (fut, rbr_out) = client.do_exchange_batches(
                    unbound_expr_exchanger.command, rbr_in
                )
                if do_instrument_reader:
                    rbr_out = instrument_reader(rbr_out, "output: ")

                # HAK: account for https://github.com/apache/arrow-rs/issues/6471
                rbr_out = copy_rbr_batches(rbr_out)
                yield from rbr_out

        gen = inner(self)
        schema = self.schema.to_pyarrow()
        return pa.RecordBatchReader.from_batches(schema, gen)

    def serve(self, make_server=None, **kwargs):
        return flight_serve_unbound(
            self.unbound_expr, make_server=make_server, **kwargs
        )


def flight_serve(
    expr,
    make_server=None,
    **kwargs,
):
    return flight_serve_unbound(expr.unbind(), make_server=make_server, **kwargs)


def flight_serve_unbound(
    unbound_expr,
    make_server=None,
    **kwargs,
):
    from xorq.flight import FlightServer
    from xorq.flight.exchanger import (
        UnboundExprExchanger,
    )

    @toolz.curry
    def do_exchange(server, command, expr):
        (_, rbr_out) = server.client.do_exchange(command, expr)
        return rbr_out

    unbound_expr_exchanger = UnboundExprExchanger(unbound_expr)
    server = (make_server or FlightServer)(**kwargs)
    server.exchangers += (unbound_expr_exchanger,)
    server.serve()
    return server, do_exchange(server, unbound_expr_exchanger.command)


@toolz.curry
def flight_expr(
    expr,
    unbound_expr,
    name=None,
    make_server=None,
    make_connection=None,
    inner_name=None,
    con=None,
    **kwargs,
):
    return (
        FlightExpr.from_exprs(
            expr,
            unbound_expr,
            make_server=make_server,
            make_connection=make_connection,
            name=inner_name,
            **kwargs,
        )
        .to_expr()
        .into_backend(con=con or expr._find_backend(), name=name)
    )


class FlightUDXF(DatabaseTableView):
    input_expr: Expr = None
    # FIXME: fix circular import issue so we can possibly pass an instance of AbstractExchanger
    udxf: type = None
    make_server: Callable = None
    make_connection: Callable = None
    do_instrument_reader: bool = False

    @classmethod
    def validate_schema(cls, input_expr, udxf):
        if not udxf.schema_in_condition(input_expr.schema()):
            schema_in_required = getattr(udxf, "schema_in_required", None)
            raise ValueError(
                "Schema validation failed"
                if schema_in_required is None
                else f"Schema validation failed, expected: {schema_in_required} found: {input_expr.schema()}"
            )
        schema_out = udxf.calc_schema_out(input_expr.schema())
        return schema_out

    @classmethod
    def from_expr(
        cls,
        input_expr,
        udxf,
        make_server=None,
        make_connection=None,
        name=None,
        **kwargs,
    ):
        from xorq.common.utils.tls_utils import TLSKwargs
        from xorq.flight import FlightServer

        def make_mtls_server():
            tls_kwargs = TLSKwargs.from_common_name(verify_client=True)
            return FlightServer(verify_client=True, **tls_kwargs.server_kwargs)

        # FIXME do we need make_connection

        schema = cls.validate_schema(input_expr, udxf)
        return cls(
            name=name or gen_name(),
            schema=schema,
            source=input_expr._find_backend(),
            input_expr=input_expr,
            udxf=udxf,
            make_server=make_server or make_mtls_server,
            make_connection=make_connection or xo_connect,
            **kwargs,
        )

    def to_rbr(self, do_instrument_reader=None):
        from xorq.flight.action import AddExchangeAction

        if do_instrument_reader is None:
            do_instrument_reader = self.do_instrument_reader

        def inner(flight_udxf):
            rbr_in = flight_udxf.input_expr.to_pyarrow_batches()
            if do_instrument_reader:
                rbr_in = instrument_reader(rbr_in, "input: ")
            with flight_udxf.make_server() as server:
                client = server.client
                client.do_action(
                    AddExchangeAction.name,
                    self.udxf,
                    options=client._options,
                )
                (fut, rbr_out) = client.do_exchange_batches(self.udxf.command, rbr_in)
                if do_instrument_reader:
                    rbr_out = instrument_reader(rbr_out, "output: ")
                # HAK: account for https://github.com/apache/arrow-rs/issues/6471
                rbr_out = copy_rbr_batches(rbr_out)
                yield from rbr_out

        gen = inner(self)
        schema = self.schema.to_pyarrow()
        return pa.RecordBatchReader.from_batches(schema, gen)


@toolz.curry
def flight_udxf(
    expr,
    process_df,
    maybe_schema_in,
    maybe_schema_out,
    name=None,
    make_server=None,
    make_connection=None,
    con=None,
    inner_name=None,
    make_udxf_kwargs=(),
    **kwargs,
):
    """
    Create a User-Defined Exchange Function (UDXF) that executes a pandas DataFrame
    transformation via Apache Arrow Flight protocol.

    This function wraps a pandas-based data processing function in an Arrow Flight
    ephemeral server, enabling distributed execution of custom user-defined functions.
    The function creates a FlightUDXF operation that can be integrated into xorq
    expression pipelines for scalable data processing.

    Parameters
    ----------
    expr : Expr
        The input Ibis expression that provides data to the UDXF. This expression's
        output will be streamed to the Flight server for processing.

    process_df : callable
        A function that takes a pandas DataFrame as input and returns a transformed
        pandas DataFrame. This function defines the core transformation logic that
        will be executed on the Flight server. The function signature should be:
        `process_df(df: pd.DataFrame) -> pd.DataFrame`

    maybe_schema_in : Schema or callable
        Input schema specification. Can be either:
        - A pyarrow Schema object defining the expected input schema
        - A callable that validates the input schema and returns True/False
        Used to validate that the input expression's schema matches expectations.

    maybe_schema_out : Schema or callable
        Output schema specification. Can be either:
        - A pyarrow Schema object defining the expected output schema
        - A callable that computes the output schema from the input schema
        Used to determine the schema of the transformed data.

    name : str, optional
        Name for the resulting table in the target backend. If not provided,
        a unique name will be generated automatically.

    make_server : callable, optional
        Factory function for creating the Arrow Flight server. Defaults to
        creating an mTLS-enabled FlightServer with client verification.
        The function should return a FlightServer instance.

    make_connection : callable, optional
        Factory function for creating connections to backends. Defaults to
        `xo.connect`. Used for establishing connections during Flight operations.

    con : Backend, optional
        Target backend connection where the result will be materialized.
        If not provided, uses the backend from the input expression.

    inner_name : str, optional
        Internal name for the FlightUDXF operation. If not provided,
        a unique name will be generated.

    make_udxf_kwargs : tuple, optional
        Additional keyword arguments to pass to the UDXF creation process.
        Should be a tuple of (key, value) pairs that will be converted to
        a dictionary and passed to `make_udxf`.

    **kwargs : dict
        Additional keyword arguments passed to the FlightUDXF constructor.

    Returns
    -------
    Expr
        A Xorq expression representing the transformed data. This expression
        can be further chained with other operations or executed to materialize
        the results.

    Examples
    --------
    Basic sentiment analysis:

    >>> import pandas as pd
    >>> import xorq.api as xo
    >>> from xorq.common.utils.toolz_utils import curry
    >>>
    >>> @curry
    >>> def add_sentiment(df: pd.DataFrame, input_col, output_col):
    ...     # Simplified sentiment analysis
    ...     sentiments = df[input_col].apply(lambda x: "POSITIVE" if "good" in x.lower() else "NEGATIVE")
    ...     return df.assign(**{output_col: sentiments})
    >>>
    >>> # Define schemas
    >>> schema_in = xo.schema({"text": "string"})
    >>> schema_out = xo.schema({"text": "string", "sentiment": "string"})
    >>>
    >>> # Create the UDXF
    >>> sentiment_udxf = xo.expr.relations.flight_udxf(
    ...     process_df=add_sentiment(input_col="text", output_col="sentiment"),
    ...     maybe_schema_in=schema_in,
    ...     maybe_schema_out=schema_out,
    ...     name="SentimentAnalyzer"
    ... )
    >>>
    >>> # Apply to data
    >>> data = xo.memtable({"text": ["This is good", "This is bad"]})
    >>> result = data.pipe(sentiment_udxf).execute()

    Data fetching and processing:

    >>> @curry
    >>> def fetch_external_data(df, api_endpoint):
    ...     # Fetch additional data for each row
    ...     results = []
    ...     for _, row in df.iterrows():
    ...         # Simulate API call
    ...         enriched_data = {"id": row["id"], "enriched": f"data_for_{row['id']}"}
    ...         results.append(enriched_data)
    ...     return pd.DataFrame(results)
    >>>
    >>> fetch_udxf = xo.expr.relations.flight_udxf(
    ...     process_df=fetch_external_data(api_endpoint="https://api.example.com"),
    ...     maybe_schema_in=xo.schema({"id": "int64"}).to_pyarrow(),
    ...     maybe_schema_out=xo.schema({"id": "int64", "enriched": "string"}).to_pyarrow(),
    ...     name="DataEnricher"
    ... ) # quartodoc: +SKIP

    Notes
    -----
    - The function uses Apache Arrow Flight for efficient data transfer between
      the client and server processes
    - By default, the Flight server uses mTLS (mutual TLS) for secure communication
    - The process_df function is executed in a separate process/server, enabling
      distributed processing and isolation
    - Schema validation ensures type safety and prevents runtime errors
    - The function is curried using toolz.curry, allowing partial application
    """

    from xorq.flight.exchanger import make_udxf

    udxf = make_udxf(
        process_df,
        maybe_schema_in,
        maybe_schema_out,
        **dict(make_udxf_kwargs),
    )
    return (
        FlightUDXF.from_expr(
            input_expr=expr,
            udxf=udxf,
            make_server=make_server,
            make_connection=make_connection,
            name=inner_name,
            **kwargs,
        )
        .to_expr()
        .into_backend(con=con or expr._find_backend(), name=name)
    )


class Read(ops.DatabaseTable):
    method_name: str = None
    read_kwargs: Any = ()
    normalize_method: Callable = None

    def make_dt(self):
        method = getattr(self.source, self.method_name)
        args = tuple(v for k, v in self.read_kwargs if k in ("path", "paths"))
        kwargs = {k: v for k, v in self.read_kwargs if k not in ("path", "paths")}
        dt = method(*args, **kwargs).op()
        return dt

    def make_unbound_dt(self):
        return ops.UnboundTable(
            name=self.name,
            schema=self.schema,
        )


_count = itertools.count()


@tracer.start_as_current_span("register_and_transform_remote_tables")
def register_and_transform_remote_tables(expr, **kwargs):
    created = {}

    op = expr.op()
    graph, _ = Graph.from_bfs(op).toposort()
    counts = defaultdict(int)
    for node in graph:
        if isinstance(node, RemoteTable):
            counts[node] += 1

        if isinstance(node, Relation):
            for arg in node.__args__:
                if isinstance(arg, RemoteTable):
                    counts[arg] += 1

    if counts:
        trace.get_current_span().add_event(
            "remote_table.replace", {"counts.values": tuple(counts.values())}
        )
    batches_table = {}
    for arg, count in counts.items():
        ex = arg.remote_expr
        batches = ex.to_pyarrow_batches()
        schema = ex.as_table().schema().to_pyarrow()
        replicas = SafeTee.tee(batches, count)
        batches_table[arg] = (schema, list(replicas))

    def mark_remote_table(node):
        schema, batchess = batches_table[node]
        name = f"{node.name}_cu{next(_count)}_t{len(batchess)}"
        reader = pa.RecordBatchReader.from_batches(schema, batchess.pop())
        result = node.source.read_record_batches(
            reader,
            table_name=name,
            **kwargs,
        )
        created[name] = node.source
        return result.op()

    def replacer(node, kwargs):
        if isinstance(node, RemoteTable):
            result = mark_remote_table(node)
            batches_table[result] = batches_table.pop(node)
            node = result
        else:
            kwargs = kwargs or {}
            if isinstance(node, Relation):
                updated = {}
                for _k, v in list(kwargs.items()):
                    try:
                        if v in batches_table:
                            updated[v] = mark_remote_table(v)

                    except TypeError:  # v may not be hashable
                        continue

                if len(updated) > 0:
                    kwargs = {
                        k: recursive_update(v, updated) for k, v in kwargs.items()
                    }

            if kwargs:
                node = node.__recreate__(kwargs)

        return node

    expr = op.replace(replacer).to_expr()
    return expr, created


def render_backend(con):
    return f"{con.name}-{id(con)}"


def get_cache_params(cache):
    from xorq.caching import (
        ParquetCache,
        ParquetSnapshotCache,
        SourceCache,
        SourceSnapshotCache,
    )

    cache_repr = None, None
    match cache:
        case ParquetCache():
            cache_repr = "modification_time", True
        case ParquetSnapshotCache():
            cache_repr = "snapshot", True
        case SourceCache():
            cache_repr = "modification_time", False
        case SourceSnapshotCache():
            cache_repr = "snapshot", False
    return cache_repr + (render_backend(cache.storage.source),)


@fmt.register(CachedNode)
def _fmt_cache_node(op, schema, parent, source, cache, **kwargs):
    strategy, parquet, backend = get_cache_params(cache)
    name = f"{op.__class__.__name__}[{parent}, strategy={strategy}, parquet={parquet}, source={backend}]\n"
    return name + render_schema(schema, 1)


@fmt.register(RemoteTable)
def _fmt_remote_table(op, name, remote_expr, **kwargs):
    name = f"{op.__class__.__name__}[{remote_expr}, name={name}]\n"
    return name + render_schema(op.schema, 1)


@fmt.register(Read)
def _fmt_read(op, name, method_name, source, **kwargs):
    backend = render_backend(source)
    name = f"{op.__class__.__name__}[name={name}, method_name={method_name}, source={backend}]\n"
    return name + render_schema(op.schema, 1)


def prepare_create_table_from_expr(con, expr, **kwargs):
    from xorq.expr.api import _transform_expr

    if (expr_backend := expr._find_backend()) != con:
        raise ValueError(f"expr backend must be {con}, is {expr_backend}")
    (table, _) = _transform_expr(expr, **kwargs)
    return table



from __future__ import annotations

import functools

from attr import (
    field,
    frozen,
)
from attr.validators import (
    instance_of,
)
from opentelemetry import trace
from public import public

from xorq.caching.storage import (
    CacheStorage,
    ParquetStorage,
    ParquetTTLStorage,
    SourceStorage,
)
from xorq.caching.strategy import (
    CacheStrategy,
    ModificationTimeStrategy,
    SnapshotStrategy,
)
from xorq.common.utils.otel_utils import tracer
from xorq.config import options
from xorq.vendor.ibis.expr import types as ir


__all__ = [  # noqa: PLE0604
    "ParquetCache",
    "ParquetSnapshotCache",
    "ParquetTTLSnapshotCache",
    "SourceCache",
    "SourceSnapshotCache",
    "GCSCache",
]


@frozen
class Cache:
    strategy = field(validator=instance_of(CacheStrategy))
    storage = field(validator=instance_of(CacheStorage))
    key_prefix = field(
        validator=instance_of(str),
        factory=functools.partial(options.get, "cache.key_prefix"),
    )

    strategy_typ = None
    storage_typ = None

    def __attrs_post_init__(self):
        assert isinstance(self.strategy, self.strategy_typ)
        assert isinstance(self.storage, self.storage_typ)

    def calc_key(self, expr):
        # the order here matters: must check is_cached before calling maybe_prevent_cross_source_caching
        if expr.ls.is_cached and expr.ls.cache is self:
            expr = expr.ls.uncached_one
        expr = maybe_prevent_cross_source_caching(expr, self)
        # FIXME: let strategy solely determine key by giving it key_prefix
        return self.key_prefix + self.strategy.calc_key(expr)

    def key_exists(self, key):
        return self.storage.exists(key)

    def exists(self, expr):
        key = self.calc_key(expr)
        return self.storage.exists(key)

    def get(self, expr: ir.Expr):
        key = self.calc_key(expr)
        if not self.key_exists(key):
            raise KeyError(key)
        else:
            return self.storage.get(key)

    def put(self, expr: ir.Expr, value):
        key = self.calc_key(expr)
        if self.key_exists(key):
            raise ValueError(f"cache entry already exists for key {key!r}")
        else:
            return self.storage.put(key, value)

    @tracer.start_as_current_span("cache.set_default")
    def set_default(self, expr: ir.Expr, default):
        span = trace.get_current_span()
        key = self.calc_key(expr)
        if not self.key_exists(key):
            span.add_event("cache.miss", {"key": key})
            with tracer.start_as_current_span("cache.put") as child_span:
                child_span.add_event("cache.miss", {"key": key})
                return self.storage.put(key, default)
        else:
            span.add_event("cache.hit", {"key": key})
            return self.storage.get(key)

    def drop(self, expr: ir.Expr):
        key = self.calc_key(expr)
        if not self.key_exists(key):
            raise KeyError(key)
        else:
            self.storage.drop(key)

    def __dask_tokenize__(self):
        from xorq.common.utils.dask_normalize.dask_normalize_utils import (
            normalize_seq_with_caller,
        )

        return normalize_seq_with_caller(
            self.strategy, self.storage, self.key_prefix, caller="normalize_cache"
        )

    @classmethod
    def from_kwargs(cls, **kwargs):
        strategy = cls.strategy_typ()
        storage = cls.storage_typ(**kwargs)
        return cls(strategy=strategy, storage=storage)


@public
@frozen
class ParquetCache(Cache):
    """Cache expressions as Parquet files using a snapshot invalidation strategy.

    This storage class saves intermediate results as Parquet files in a specified
    directory and uses a snapshot-based approach for cache invalidation.
    The snapshot strategy ensures cached data is only invalidated when the
    expression's definition changes, making it suitable for stable datasets.

    Parameters
    ----------
    source : ibis.backends.BaseBackend
        The backend to use for execution. Defaults to xorq's default backend.
    path : Path
        The directory where Parquet files will be stored. Defaults to
        xorq.config.options.cache.default_path.
    """

    strategy_typ = ModificationTimeStrategy
    storage_typ = ParquetStorage


@public
@frozen
class ParquetSnapshotCache(Cache):
    """Cache expressions as Parquet files using a snapshot invalidation strategy.

    This storage class saves intermediate results as Parquet files in a specified
    directory and uses a snapshot-based approach for cache invalidation.
    The snapshot strategy ensures cached data is only invalidated when the
    expression's definition changes, making it suitable for stable datasets.

    Parameters
    ----------
    source : ibis.backends.BaseBackend
        The backend to use for execution. Defaults to xorq's default backend.
    path : Path
        The directory where Parquet files will be stored. Defaults to
        xorq.config.options.cache.default_path.
    """

    strategy_typ = SnapshotStrategy
    storage_typ = ParquetStorage


@public
@frozen
class ParquetTTLSnapshotCache(Cache):
    strategy_typ = SnapshotStrategy
    storage_typ = ParquetTTLStorage


@public
@frozen
class SourceCache(Cache):
    strategy_typ = ModificationTimeStrategy
    storage_typ = SourceStorage


@public
@frozen
class SourceSnapshotCache(Cache):
    strategy_typ = SnapshotStrategy
    storage_typ = SourceStorage


@public
@frozen
class GCSCache(Cache):
    strategy_typ = ModificationTimeStrategy
    storage_typ = None

    def __attrs_post_init__(self):
        from xorq.common.utils.gcloud_utils import GCStorage

        assert isinstance(self.strategy, self.strategy_typ)
        assert isinstance(self.storage, GCStorage)

    @classmethod
    def from_kwargs(cls, bucket_name, source):
        from xorq.common.utils.gcloud_utils import GCStorage

        strategy = cls.strategy_typ()
        storage = GCStorage(bucket_name=bucket_name, source=source)
        return cls(strategy=strategy, storage=storage)


def maybe_prevent_cross_source_caching(expr, storage):
    from xorq.expr.relations import (
        into_backend,
    )

    if storage.storage.source != expr._find_backend():
        expr = into_backend(expr, storage.storage.source)
    return expr



"""Tests for YAML build serialization of sklearn pipeline expressions."""

import pandas as pd
import pandas.testing as tm
import pyarrow as pa
import pyarrow.parquet as pq
import pytest

import xorq.api as xo
import xorq.ibis_yaml.translate as translate_mod
from xorq.expr.ml.cross_validation import _make_folds_from_sklearn
from xorq.ibis_yaml.common import FROM_YAML_HANDLERS
from xorq.ibis_yaml.compiler import build_expr, load_expr


sklearn = pytest.importorskip("sklearn")

from sklearn.datasets import make_regression  # noqa: E402
from sklearn.model_selection import TimeSeriesSplit  # noqa: E402


@pytest.fixture(scope="module")
def lasso_timeseries_fold_expr(tmp_path_factory):
    """Build a TimeSeriesSplit fold expression using deferred_read_parquet.

    Writes regression data to a parquet file so the expression graph is
    YAML-serializable via the Read op handler in the ibis_yaml compiler.
    _make_folds_from_sklearn is purely deferred — no data execution required.
    """
    X, y = make_regression(
        n_samples=200, n_features=4, n_informative=3, random_state=42
    )
    df = pd.DataFrame(X, columns=["f0", "f1", "f2", "f3"]).assign(
        target=y, t=range(200)
    )
    parquet_dir = tmp_path_factory.mktemp("data")
    parquet_path = parquet_dir / "regression_data.parquet"
    pq.write_table(pa.Table.from_pandas(df, preserve_index=False), parquet_path)

    expr = xo.deferred_read_parquet(parquet_path)
    features = ("f0", "f1", "f2", "f3")
    target = "target"

    return _make_folds_from_sklearn(
        expr,
        TimeSeriesSplit(n_splits=3),
        features,
        target,
        random_seed=42,
        order_by="t",
    )


def test_lasso_timeseries_cv_bytes_to_yaml_invoked(
    lasso_timeseries_fold_expr, tmp_path, monkeypatch
):
    """_bytes_to_yaml is called when building a TimeSeriesSplit fold expression.

    _make_folds_from_sklearn calls _make_fold_udwf, which pickles the splitter as
    cv_bytes and stores it in the UDWF config FrozenDict. Building the fold
    expression to YAML must invoke _bytes_to_yaml at least once for that bytes value.
    """
    original = translate_mod._bytes_to_yaml
    call_count = {"n": 0}

    def spy(*args, **kwargs):
        call_count["n"] += 1
        return original(*args, **kwargs)

    # Register spy into the singledispatch registry; restore original in finally.
    translate_mod.translate_to_yaml.register(bytes)(spy)
    try:
        build_expr(lasso_timeseries_fold_expr, builds_dir=tmp_path)
    finally:
        translate_mod.translate_to_yaml.register(bytes)(original)

    assert call_count["n"] > 0, (
        "_bytes_to_yaml was never called during YAML serialization. "
        "Check that it is registered via @translate_to_yaml.register(bytes)."
    )


def test_lasso_timeseries_cv_bytes_from_yaml_invoked(
    lasso_timeseries_fold_expr, tmp_path, monkeypatch
):
    """_bytes_from_yaml is called when loading a TimeSeriesSplit fold expression.

    After building the fold expression to YAML, loading it must invoke
    _bytes_from_yaml at least once to reconstruct the pickled splitter bytes.
    """
    expr_path = build_expr(lasso_timeseries_fold_expr, builds_dir=tmp_path)

    original = translate_mod._bytes_from_yaml
    call_count = {"n": 0}

    def spy(*args, **kwargs):
        call_count["n"] += 1
        return original(*args, **kwargs)

    monkeypatch.setitem(FROM_YAML_HANDLERS, "bytes", spy)

    load_expr(expr_path)

    assert call_count["n"] > 0, (
        "_bytes_from_yaml was never called during YAML deserialization. "
        "Check that it is registered via @register_from_yaml_handler('bytes')."
    )


def test_lasso_timeseries_cv_yaml_roundtrip(lasso_timeseries_fold_expr, tmp_path):
    """The fold expression roundtrips through build/load with identical results."""
    expr_path = build_expr(lasso_timeseries_fold_expr, builds_dir=tmp_path)
    roundtrip_expr = load_expr(expr_path)

    sort_cols = lasso_timeseries_fold_expr.columns
    original_df = (
        lasso_timeseries_fold_expr.execute()
        .sort_values(sort_cols)
        .reset_index(drop=True)
    )
    roundtrip_df = (
        roundtrip_expr.execute().sort_values(sort_cols).reset_index(drop=True)
    )
    tm.assert_frame_equal(original_df, roundtrip_df)



import gcsfs
import pyarrow as pa
import pyarrow.parquet as pq
from attr import (
    field,
    frozen,
)
from attr.validators import (
    instance_of,
)
from google.cloud import storage
from toolz import curry

from xorq.caching.storage import CacheStorage
from xorq.config import _backend_init
from xorq.vendor.ibis.backends import BaseBackend


@curry
def rbr_from_fs(fs, path):
    def get_schema(fs, path):
        with fs.open(path, "rb") as fh:
            pf = pq.ParquetFile(fh)
            schema = pf.schema.to_arrow_schema()
            return schema

    def gen_batches(fs, path):
        with fs.open(path, "rb") as fh:
            pf = pq.ParquetFile(fh)
            yield from pf.iter_batches()

    rbr = pa.RecordBatchReader.from_batches(
        get_schema(fs, path),
        gen_batches(fs, path),
    )
    return rbr


@curry
def rbr_to_fs(fs, path, rbr, **kwargs):
    with fs.open(path, "wb") as fh:
        with pq.ParquetWriter(fh, rbr.schema, **kwargs) as writer:
            for batch in rbr:
                writer.write_batch(batch)


@frozen
class GCStorage(CacheStorage):
    bucket_name: str = field(validator=instance_of(str))
    source = field(
        validator=instance_of(BaseBackend),
        factory=_backend_init,
    )
    fs: gcsfs.GCSFileSystem = field(init=False)

    def __attrs_post_init__(self):
        assert hasattr(self.source, "read_record_batches")
        object.__setattr__(self, "fs", gcsfs.GCSFileSystem())

    def __dask_tokenize__(self):
        from xorq.common.utils.dask_normalize.dask_normalize_utils import (
            normalize_seq_with_caller,
        )

        return normalize_seq_with_caller(
            self.source, self.bucket_name, caller="normalize_gc_storage"
        )

    def get_path(self, key):
        path = f"{self.bucket_name}/{key}.parquet"
        return path

    def exists(self, key):
        path = self.get_path(key)
        return self.fs.exists(path)

    def get(self, key):
        path = self.get_path(key)
        rbr = rbr_from_fs(self.fs, path)
        op = self.source.read_record_batches(rbr).op()
        return op

    def put(self, key, value):
        path = self.get_path(key)
        rbr = value.to_expr().to_pyarrow_batches()
        rbr_to_fs(self.fs, path, rbr)
        return self.get(key)

    def drop(self, key):
        path = self.get_path(key)
        self.fs.delete(path)


def get_file_metadata(uri, client=None):
    blob = storage.Blob.from_string(uri)
    # Refresh metadata (required for accurate timestamps)
    blob.reload(client or storage.Client.create_anonymous_client())

    # Extract relevant metadata
    metadata = tuple(
        (name, getattr(blob, name))
        for name in (
            "content_type",
            "updated",
            "size",
        )
    )

    return metadata



"""Convenient column selectors.

::: {.callout-tip}
## Check out the [blog post on selectors](../posts/selectors) for examples!
:::

## Rationale

Column selectors are convenience functions for selecting columns that share some property.

## Discussion

For example, a common task is to be able to select all numeric columns for a
subsequent computation.

Without selectors this becomes quite verbose and tedious to write:

>>> import ibis
>>> t = ibis.table(dict(a="int", b="string", c="array<int>", abcd="float"))
>>> expr = t.select([t[c] for c in t.columns if t[c].type().is_numeric()])
>>> expr.columns
['a', 'abcd']

Compare that to the [`numeric`](#ibis.selectors.numeric) selector:

>>> import ibis.selectors as s
>>> expr = t.select(s.numeric())
>>> expr.columns
['a', 'abcd']

When there are multiple properties to check it gets worse:

>>> expr = t.select(
...     [
...         t[c]
...         for c in t.columns
...         if t[c].type().is_numeric() or t[c].type().is_string()
...         if ("a" in c or "b" in c or "cd" in c)
...     ]
... )
>>> expr.columns
['a', 'b', 'abcd']

Using a composition of selectors this is much less tiresome:

>>> expr = t.select((s.numeric() | s.of_type("string")) & s.contains(("a", "b", "cd")))
>>> expr.columns
['a', 'b', 'abcd']
"""

from __future__ import annotations

import builtins
import inspect
import operator
import re
from collections.abc import Callable, Iterable, Mapping, Sequence
from functools import reduce
from typing import Optional, Union

from public import public

import xorq.common.exceptions as exc
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
import xorq.vendor.ibis.expr.types as ir
from xorq.vendor.ibis import util
from xorq.vendor.ibis.common.collections import frozendict  # noqa: TCH001
from xorq.vendor.ibis.common.deferred import Deferred, Resolver
from xorq.vendor.ibis.common.grounds import Concrete, Singleton
from xorq.vendor.ibis.common.selectors import All, Any, Expandable, Selector
from xorq.vendor.ibis.common.typing import VarTuple  # noqa: TCH001


def __getattr__(name):
    if name == "c":
        util.warn_deprecated(
            "c", instead="use `ibis.selectors.cols` instead", as_of="9.5"
        )
        return cols
    elif name == "r":
        util.warn_deprecated(
            "r", instead="use `ibis.selectors.index` instead", as_of="9.5"
        )
        return index
    raise AttributeError(name)


class Where(Selector):
    predicate: Callable[[ir.Value], bool]

    def expand_names(self, table: ir.Table) -> frozenset[str]:
        predicate = self.predicate
        return frozenset(col for col in table.columns if predicate(table[col]))


@public
def where(predicate: Callable[[ir.Value], bool]) -> Selector:
    """Select columns that satisfy `predicate`.

    Use this selector when one of the other selectors does not meet your needs.

    Parameters
    ----------
    predicate
        A callable that accepts an ibis value expression and returns a `bool`

    Examples
    --------
    >>> import ibis
    >>> import ibis.selectors as s
    >>> t = ibis.table(dict(a="float32"), name="t")
    >>> expr = t.select(s.where(lambda col: col.get_name() == "a"))
    >>> expr.columns
    ['a']

    """
    return Where(predicate)


@public
def numeric() -> Selector:
    """Return numeric columns.

    Examples
    --------
    >>> import ibis
    >>> import ibis.selectors as s
    >>> t = ibis.table(dict(a="int", b="string", c="array<string>"), name="t")
    >>> t.columns
    ['a', 'b', 'c']
    >>> expr = t.select(s.numeric())  # `a` has integer type, so it's numeric
    >>> expr.columns
    ['a']

    See Also
    --------
    [`of_type`](#ibis.selectors.of_type)

    """
    return of_type(dt.Numeric)


class OfType(Selector):
    predicate: Callable[[dt.DataType], bool]

    def expand_names(self, table: ir.Table) -> frozenset[str]:
        predicate = self.predicate
        return frozenset(name for name, typ in table.schema().items() if predicate(typ))


@public
def of_type(dtype: dt.DataType | str | type[dt.DataType]) -> Selector:
    """Select columns of type `dtype`.

    Parameters
    ----------
    dtype
        `DataType` instance, `str` or `DataType` class

    Examples
    --------
    Select according to a specific `DataType` instance

    >>> import ibis
    >>> import ibis.expr.datatypes as dt
    >>> import ibis.selectors as s
    >>> t = ibis.table(dict(name="string", siblings="array<string>", parents="array<int64>"))
    >>> expr = t.select(s.of_type(dt.Array(dt.string)))
    >>> expr.columns
    ['siblings']

    Strings are also accepted

    >>> expr = t.select(s.of_type("array<string>"))
    >>> expr.columns
    ['siblings']

    Abstract/unparametrized types may also be specified by their string name
    (e.g. "integer" for any integer type), or by passing in a `DataType` class
    instead. The following options are equivalent.

    >>> expr1 = t.select(s.of_type("array"))
    >>> expr2 = t.select(s.of_type(dt.Array))
    >>> expr1.equals(expr2)
    True
    >>> expr2.columns
    ['siblings', 'parents']

    See Also
    --------
    [`numeric`](#ibis.selectors.numeric)

    """
    if isinstance(dtype, str):
        # A mapping of abstract or parametric types, to allow selecting all
        # subclasses/parametrizations of these types, rather than only a
        # specific instance.
        abstract = {
            "array": dt.Array,
            "decimal": dt.Decimal,
            "floating": dt.Floating,
            "geospatial": dt.GeoSpatial,
            "integer": dt.Integer,
            "map": dt.Map,
            "numeric": dt.Numeric,
            "struct": dt.Struct,
            "temporal": dt.Temporal,
        }

        if dtype_cls := abstract.get(dtype.lower()):
            predicate = lambda typ, dtype_cls=dtype_cls: isinstance(typ, dtype_cls)  # noqa: E731
        else:
            dtype = dt.dtype(dtype)
            predicate = lambda typ, dtype=dtype: typ == dtype  # noqa: E731

    elif inspect.isclass(dtype) and issubclass(dtype, dt.DataType):
        predicate = lambda typ, dtype_cls=dtype: isinstance(typ, dtype_cls)  # noqa: E731
    else:
        dtype = dt.dtype(dtype)
        predicate = lambda typ, dtype=dtype: typ == dtype  # noqa: E731

    return OfType(predicate)


class StartsWith(Selector):
    prefixes: str | VarTuple[str]

    def expand_names(self, table: ir.Table) -> frozenset[str]:
        prefixes = self.prefixes
        return frozenset(col for col in table.columns if col.startswith(prefixes))


@public
def startswith(prefixes: str | tuple[str, ...]) -> Selector:
    """Select columns whose name starts with one of `prefixes`.

    Parameters
    ----------
    prefixes
        Prefixes to compare column names against

    Examples
    --------
    >>> import ibis
    >>> import ibis.selectors as s
    >>> t = ibis.table(dict(apples="int", oranges="float", bananas="bool"), name="t")
    >>> expr = t.select(s.startswith(("a", "b")))
    >>> expr.columns
    ['apples', 'bananas']

    See Also
    --------
    [`endswith`](#ibis.selectors.endswith)

    """
    return StartsWith(prefixes)


class EndsWith(Selector):
    suffixes: str | VarTuple[str]

    def expand_names(self, table: ir.Table) -> frozenset[str]:
        suffixes = self.suffixes
        return frozenset(col for col in table.columns if col.endswith(suffixes))


@public
def endswith(suffixes: str | tuple[str, ...]) -> Selector:
    """Select columns whose name ends with one of `suffixes`.

    Parameters
    ----------
    suffixes
        Suffixes to compare column names against

    See Also
    --------
    [`startswith`](#ibis.selectors.startswith)

    """
    return EndsWith(suffixes)


class Contains(Selector):
    needles: VarTuple[str]
    how: Callable[[Iterable[bool]], bool]

    def expand_names(self, table: ir.Table) -> frozenset[str]:
        needles = self.needles
        how = self.how
        return frozenset(
            col for col in table.columns if how(map(col.__contains__, needles))
        )


@public
def contains(
    needles: str | tuple[str, ...],
    how: Callable[[Iterable[bool]], bool] = builtins.any,
) -> Selector:
    """Return columns whose name contains `needles`.

    Parameters
    ----------
    needles
        One or more strings to search for in column names
    how
        A boolean reduction to allow the configuration of how `needles` are summarized.

    Examples
    --------
    Select columns that contain either `"a"` or `"b"`

    >>> import ibis
    >>> import ibis.selectors as s
    >>> t = ibis.table(
    ...     dict(a="int64", b="string", c="float", d="array<int16>", ab="struct<x: int>")
    ... )
    >>> expr = t.select(s.contains(("a", "b")))
    >>> expr.columns
    ['a', 'b', 'ab']

    Select columns that contain all of `"a"` and `"b"`, that is, both `"a"` and
    `"b"` must be in each column's name to match.

    >>> expr = t.select(s.contains(("a", "b"), how=all))
    >>> expr.columns
    ['ab']

    See Also
    --------
    [`matches`](#ibis.selectors.matches)

    """
    return Contains(tuple(util.promote_list(needles)), how=how)


class Matches(Selector):
    regex: re.Pattern

    def expand_names(self, table: ir.Table) -> frozenset[str]:
        return frozenset(filter(self.regex.search, table.columns))


@public
def matches(regex: str | re.Pattern) -> Selector:
    """Return columns whose name matches the regular expression `regex`.

    Parameters
    ----------
    regex
        A string or `re.Pattern` object

    Examples
    --------
    >>> import ibis
    >>> import ibis.selectors as s
    >>> t = ibis.table(dict(ab="string", abd="int", be="array<string>"))
    >>> expr = t.select(s.matches(r"ab+"))
    >>> expr.columns
    ['ab', 'abd']

    See Also
    --------
    [`contains`](#ibis.selectors.contains)

    """
    return Matches(re.compile(regex))


@public
def any_of(*predicates: str | Selector) -> Selector:
    """Include columns satisfying any of `predicates`."""
    return Any(tuple(map(_to_selector, predicates)))


@public
def all_of(*predicates: str | Selector) -> Selector:
    """Include columns satisfying all of `predicates`."""
    return All(tuple(map(_to_selector, predicates)))


class Cols(Selector):
    names: frozenset[str]

    def expand_names(self, table: ir.Table) -> frozenset[str]:
        names = self.names
        columns = table.columns
        if extra_cols := sorted(names.difference(columns)):
            raise exc.XorqInputError(
                f"Columns {extra_cols} are not present in {columns}"
            )
        return names


@public
def cols(*names: str | ir.Column) -> Selector:
    """Select specific column names.

    Parameters
    ----------
    names
        The column names to select

    Examples
    --------
    >>> import ibis
    >>> import ibis.selectors as s
    >>> t = ibis.table({"a": "int", "b": "int", "c": "int"})
    >>> expr = t.select(s.cols("a", "b"))
    >>> expr.columns
    ['a', 'b']

    See Also
    --------
    [`index`](#ibis.selectors.cols)
    """
    names = frozenset(col if isinstance(col, str) else col.get_name() for col in names)
    return Cols(names)


class Across(Concrete, Expandable):
    selector: Selector
    funcs: Union[
        Resolver,
        Callable[[ir.Value], ir.Value],
        frozendict[Optional[str], Union[Resolver, Callable[[ir.Value], ir.Value]]],
    ]
    names: Union[str, Callable[[str, Optional[str]], str]]

    def expand(self, table: ir.Table) -> Sequence[ir.Value]:
        expanded = []

        names = self.names
        cols = self.selector.expand(table)
        for func_name, func in self.funcs.items():
            for orig_col in cols:
                if isinstance(func, Resolver):
                    col = func.resolve({"_": orig_col})
                else:
                    col = func(orig_col)

                orig_name = orig_col.get_name()

                if callable(names):
                    name = names(orig_name, func_name)
                else:
                    name = names.format(col=orig_name, fn=func_name)

                if not isinstance(col.op(), ops.Alias):
                    col = col.name(name)

                expanded.append(col)

        return expanded


@public
def across(
    selector: Selector | Iterable[str] | str,
    func: Deferred
    | Callable[[ir.Value], ir.Value]
    | Mapping[str | None, Deferred | Callable[[ir.Value], ir.Value]],
    names: str | Callable[[str, str | None], str] | None = None,
) -> Across:
    """Apply data transformations across multiple columns.

    Parameters
    ----------
    selector
        An expression that selects columns on which the transformation function
        will be applied, an iterable of `str` column names or a single `str`
        column name.
    func
        A function (or dictionary of functions) to use to transform the data.
    names
        A lambda function or a format string to name the columns created by the
        transformation function.

    Returns
    -------
    Across
        An `Across` selector object

    Examples
    --------
    >>> import ibis
    >>> ibis.options.interactive = True
    >>> from ibis import _, selectors as s
    >>> t = ibis.examples.penguins.fetch()
    >>> t.select(s.startswith("bill")).mutate(
    ...     s.across(s.numeric(), dict(centered=_ - _.mean()), names="{fn}_{col}")
    ... )
    ┏━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━┓
    ┃ bill_length_mm ┃ bill_depth_mm ┃ centered_bill_length_mm ┃ … ┃
    ┡━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━╇━━━┩
    │ float64        │ float64       │ float64                 │ … │
    ├────────────────┼───────────────┼─────────────────────────┼───┤
    │           39.1 │          18.7 │                -4.82193 │ … │
    │           39.5 │          17.4 │                -4.42193 │ … │
    │           40.3 │          18.0 │                -3.62193 │ … │
    │           NULL │          NULL │                    NULL │ … │
    │           36.7 │          19.3 │                -7.22193 │ … │
    │           39.3 │          20.6 │                -4.62193 │ … │
    │           38.9 │          17.8 │                -5.02193 │ … │
    │           39.2 │          19.6 │                -4.72193 │ … │
    │           34.1 │          18.1 │                -9.82193 │ … │
    │           42.0 │          20.2 │                -1.92193 │ … │
    │              … │             … │                       … │ … │
    └────────────────┴───────────────┴─────────────────────────┴───┘

    """
    if names is None:
        names = lambda col, fn: "_".join(filter(None, (col, fn)))  # noqa: E731
    funcs = dict(func if isinstance(func, Mapping) else {None: func})
    selector = _to_selector(selector)
    return Across(selector=selector, funcs=funcs, names=names)


class IfAnyAll(Concrete, Expandable):
    selector: Selector
    predicate: Union[Resolver, Callable[[ir.Value], ir.BooleanValue]]
    summarizer: Callable[[ir.BooleanValue, ir.BooleanValue], ir.BooleanValue]

    def expand(self, table: ir.Table) -> Sequence[ir.Value]:
        func = self.predicate

        if isinstance(func, Resolver):
            fn = lambda col, func=func: func.resolve({"_": col})  # noqa: E731
        else:
            fn = func

        return [reduce(self.summarizer, map(fn, self.selector.expand(table)))]


@public
def if_any(selector: Selector, predicate: Deferred | Callable) -> IfAnyAll:
    """Return the **disjunction** of `predicate` applied on all `selector` columns.

    Parameters
    ----------
    selector
        A column selector
    predicate
        A callable or deferred object defining a predicate to apply to each
        column from `selector`.

    Examples
    --------
    >>> import ibis
    >>> from xorq.vendor.ibis import selectors as s, _
    >>> ibis.options.interactive = True
    >>> penguins = ibis.examples.penguins.fetch()
    >>> cols = s.across(s.endswith("_mm"), (_ - _.mean()) / _.std())
    >>> expr = penguins.mutate(cols).filter(s.if_any(s.endswith("_mm"), _.abs() > 2))
    >>> expr_by_hand = penguins.mutate(cols).filter(
    ...     (_.bill_length_mm.abs() > 2)
    ...     | (_.bill_depth_mm.abs() > 2)
    ...     | (_.flipper_length_mm.abs() > 2)
    ... )
    >>> expr.equals(expr_by_hand)
    True
    >>> expr
    ┏━━━━━━━━━┳━━━━━━━━┳━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━┳━━━┓
    ┃ species ┃ island ┃ bill_length_mm ┃ bill_depth_mm ┃ flipper_length_mm ┃ … ┃
    ┡━━━━━━━━━╇━━━━━━━━╇━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━╇━━━┩
    │ string  │ string │ float64        │ float64       │ float64           │ … │
    ├─────────┼────────┼────────────────┼───────────────┼───────────────────┼───┤
    │ Adelie  │ Biscoe │      -1.103002 │      0.733662 │         -2.056307 │ … │
    │ Gentoo  │ Biscoe │       1.113285 │     -0.431017 │          2.068368 │ … │
    │ Gentoo  │ Biscoe │       2.871660 │     -0.076550 │          2.068368 │ … │
    │ Gentoo  │ Biscoe │       1.900890 │     -0.734846 │          2.139483 │ … │
    │ Gentoo  │ Biscoe │       1.076652 │     -0.177826 │          2.068368 │ … │
    │ Gentoo  │ Biscoe │       0.856855 │     -0.582932 │          2.068368 │ … │
    │ Gentoo  │ Biscoe │       1.497929 │     -0.076550 │          2.068368 │ … │
    │ Gentoo  │ Biscoe │       1.388031 │     -0.431017 │          2.068368 │ … │
    │ Gentoo  │ Biscoe │       2.047422 │     -0.582932 │          2.068368 │ … │
    │ Adelie  │ Dream  │      -2.165354 │     -0.836123 │         -0.918466 │ … │
    │ …       │ …      │              … │             … │                 … │ … │
    └─────────┴────────┴────────────────┴───────────────┴───────────────────┴───┘

    """
    return IfAnyAll(selector=selector, predicate=predicate, summarizer=operator.or_)


@public
def if_all(selector: Selector, predicate: Deferred | Callable) -> IfAnyAll:
    """Return the **conjunction** of `predicate` applied on all `selector` columns.

    Parameters
    ----------
    selector
        A column selector
    predicate
        A callable or deferred object defining a predicate to apply to each
        column from `selector`.

    Examples
    --------
    >>> import ibis
    >>> from xorq.vendor.ibis import selectors as s, _
    >>> ibis.options.interactive = True
    >>> penguins = ibis.examples.penguins.fetch()
    >>> cols = s.across(s.endswith("_mm"), (_ - _.mean()) / _.std())
    >>> expr = penguins.mutate(cols).filter(s.if_all(s.endswith("_mm"), _.abs() > 1))
    >>> expr_by_hand = penguins.mutate(cols).filter(
    ...     (_.bill_length_mm.abs() > 1)
    ...     & (_.bill_depth_mm.abs() > 1)
    ...     & (_.flipper_length_mm.abs() > 1)
    ... )
    >>> expr.equals(expr_by_hand)
    True
    >>> expr
    ┏━━━━━━━━━┳━━━━━━━━━━━┳━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━┳━━━┓
    ┃ species ┃ island    ┃ bill_length_mm ┃ bill_depth_mm ┃ flipper_length_mm ┃ … ┃
    ┡━━━━━━━━━╇━━━━━━━━━━━╇━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━╇━━━┩
    │ string  │ string    │ float64        │ float64       │ float64           │ … │
    ├─────────┼───────────┼────────────────┼───────────────┼───────────────────┼───┤
    │ Adelie  │ Dream     │      -1.157951 │      1.088129 │         -1.416272 │ … │
    │ Adelie  │ Torgersen │      -1.231217 │      1.138768 │         -1.202926 │ … │
    │ Gentoo  │ Biscoe    │       1.149917 │     -1.443781 │          1.214987 │ … │
    │ Gentoo  │ Biscoe    │       1.040019 │     -1.089314 │          1.072757 │ … │
    │ Gentoo  │ Biscoe    │       1.131601 │     -1.089314 │          1.712792 │ … │
    │ Gentoo  │ Biscoe    │       1.241499 │     -1.089314 │          1.570562 │ … │
    │ Gentoo  │ Biscoe    │       1.351398 │     -1.494420 │          1.214987 │ … │
    └─────────┴───────────┴────────────────┴───────────────┴───────────────────┴───┘

    """
    return IfAnyAll(selector=selector, predicate=predicate, summarizer=operator.and_)


class Slice(Concrete):
    """Hashable and smaller-scoped slice object versus the builtin one."""

    start: int | str | None = None
    stop: int | str | None = None
    step: int | None = None


class ColumnIndex(Selector):
    key: str | int | Slice | VarTuple[int | str]

    @staticmethod
    def slice_key_to_int(
        value: int | str | None, name_locs: Mapping[str, int], offset: int
    ) -> int:
        if value is None or isinstance(value, int):
            return value
        else:
            assert isinstance(value, str), f"expected `str` got {type(value)}"
            return name_locs[value] + offset

    def expand_names(self, table: ir.Table) -> frozenset[str]:
        name_locs = table.schema()._name_locs
        key = self.key

        if isinstance(key, str):
            iterable = (key,)
        elif isinstance(key, int):
            iterable = (table.columns[key],)
        elif isinstance(key, Slice):
            start = self.slice_key_to_int(key.start, name_locs, offset=0)
            stop = self.slice_key_to_int(key.stop, name_locs, offset=1)
            step = key.step
            iterable = table.columns[start:stop:step]
        else:
            iterable = (
                table.columns[el if isinstance(el, int) else name_locs[el]]
                for el in key
            )
        return frozenset(iterable)


class Indexable(Singleton):
    def __getitem__(self, key: str | int | slice | Iterable[int | str]):
        if isinstance(key, slice):
            key = Slice(key.start, key.stop, key.step)
        return ColumnIndex(key)


index = Indexable()
"""Select columns by index.

Examples
--------
>>> import ibis
>>> import ibis.selectors as s
>>> t = ibis.table(
...     {"a": "int", "b": "int", "c": "int", "d": "int", "e": "int"}
... )

Select one column by numeric index:
>>> expr = t.select(s.index[0])
>>> expr.columns
['a']

Select multiple columns by numeric index:
>>> expr = t.select(s.index[[0, 1]])
>>> expr.columns
['a', 'b']

Select a slice of columns by numeric index:
>>> expr = t.select(s.index[1:4])
>>> expr.columns
['b', 'c', 'd']

Select a slice of columns by name:
>>> expr = t.select(s.index["b":"d"])
>>> expr.columns
['b', 'c', 'd']

See Also
--------
[`cols`](#ibis.selectors.cols)
"""


class First(Singleton, Selector):
    def expand(self, table: ir.Table) -> Sequence[ir.Value]:
        return [table[0]]

    def expand_names(self, table: ir.Table) -> frozenset[str]:
        return frozenset((table.columns[0],))


@public
def first() -> Selector:
    """Return the first column of a table."""
    return First()


class Last(Singleton, Selector):
    def expand(self, table: ir.Table) -> Sequence[ir.Value]:
        return [table[-1]]

    def expand_names(self, table: ir.Table) -> frozenset[str]:
        return frozenset((table.columns[-1],))


@public
def last() -> Selector:
    """Return the last column of a table."""
    return Last()


class AllColumns(Singleton, Selector):
    def expand(self, table: ir.Table) -> Sequence[ir.Value]:
        return list(map(table.__getitem__, table.columns))

    def expand_names(self, table: ir.Table) -> frozenset[str]:
        return frozenset(table.columns)


@public
def all() -> Selector:
    """Return every column from a table."""
    return AllColumns()


class NoColumns(Singleton, Selector):
    def expand(self, table: ir.Table) -> Sequence[ir.Value]:
        return []

    def expand_names(self, table: ir.Table) -> frozenset[str]:
        return frozenset()


@public
def none() -> Selector:
    """Return no columns."""
    return NoColumns()


def _to_selector(
    obj: str | Selector | ir.Column | Sequence[str | Selector | ir.Column],
) -> Selector:
    """Convert an object to a `Selector`."""
    if isinstance(obj, Selector):
        return obj
    elif isinstance(obj, ir.Column):
        return cols(obj.get_name())
    elif isinstance(obj, str):
        return cols(obj)
    elif isinstance(obj, Expandable):
        raise exc.XorqInputError(
            f"Cannot compose {obj.__class__.__name__} with other selectors"
        )
    elif not obj:
        return none()
    else:
        return any_of(*obj)



from __future__ import annotations

import gzip

import pandas as pd
import pyarrow as pa
import pyarrow.dataset as ds
import pytest

import xorq.api as xo
from xorq.caching import ParquetCache
from xorq.tests.util import assert_frame_equal
from xorq.vendor.ibis import Schema
from xorq.vendor.ibis.common.collections import FrozenDict


@pytest.fixture
def gzip_csv(data_dir, tmp_path):
    basename = "diamonds.csv"
    f = tmp_path.joinpath(f"{basename}.gz")
    data = data_dir.joinpath("csv", basename).read_bytes()
    f.write_bytes(gzip.compress(data))
    return str(f.absolute())


@pytest.fixture
def ctx(con):
    return con.con


def test_register_csv(con, data_dir):
    fname = "diamonds.csv"
    table_name = "diamonds"
    table = con.read_csv(data_dir / "csv" / fname, table_name=table_name)
    assert any(table_name in t for t in con.list_tables())
    assert table.count().execute() > 0


@pytest.mark.xfail
def test_register_csv_gz(con, data_dir, gzip_csv):
    # TODO review why there is a regression
    table = con.read_csv(gzip_csv, table_name="diamonds")
    assert table.count().execute() > 0


def test_register_with_dotted_name(con, data_dir, tmp_path):
    basename = "foo.bar.baz/diamonds.csv"
    f = tmp_path.joinpath(basename)
    f.parent.mkdir()
    data = data_dir.joinpath("csv", "diamonds.csv").read_bytes()
    f.write_bytes(data)
    table = con.read_csv(str(f.absolute()), table_name="diamonds")
    assert table.count().execute() > 0


def test_register_parquet(con, data_dir):
    fname = "functional_alltypes.parquet"
    table_name = "funk_all"
    table = con.read_parquet(data_dir / "parquet" / fname, table_name=table_name)

    assert any(table_name in t for t in con.list_tables())
    assert table.count().execute() > 0


def test_read_parquet_with_custom_file_extension(con, tmp_path):
    parquet_path = (tmp_path / "data").with_suffix(".arquet")

    expected = pd.DataFrame([{"maxitem": 43182839, "n": 1000}])
    expected.to_parquet(parquet_path)

    actual = xo.read_parquet(parquet_path).execute()

    assert_frame_equal(expected, actual)


def test_read_csv_with_custom_file_extension(con, tmp_path):
    csv_path = (tmp_path / "data").with_suffix(".acsv")

    expected = pd.DataFrame([{"maxitem": 43182839, "n": 1000}])
    expected.to_csv(csv_path, index=False)

    actual = xo.read_csv(csv_path).execute()

    assert_frame_equal(expected, actual)


def test_read_csv(con, data_dir):
    t = con.read_csv(data_dir / "csv" / "functional_alltypes.csv")
    assert t.count().execute()


def test_read_csv_from_url(con):
    t = con.read_csv(
        "https://raw.githubusercontent.com/ibis-project/testing-data/refs/heads/master/csv/astronauts.csv"
    )
    assert t.head().execute() is not None


@pytest.mark.s3
def test_read_csv_from_s3(con):
    schema = pa.schema(
        [
            pa.field("question", pa.string()),
            pa.field("product_description", pa.string()),
            pa.field("image_url", pa.string()),
            pa.field("label", pa.uint8(), nullable=True),
        ]
    )

    t = con.read_csv(
        "s3://humor-detection-pds/Humorous.csv",
        schema=schema,
        storage_options={
            "aws.region": "us-west-2",
        },
    )
    assert t.head().execute() is not None


def test_read_parquet(con, data_dir):
    t = con.read_parquet(data_dir / "parquet" / "functional_alltypes.parquet")
    assert t.count().execute()


def test_read_parquet_from_url(con):
    t = con.read_parquet(
        "https://nasa-avionics-data-ml.s3.us-east-2.amazonaws.com/Tail_652_1_parquet/652200101120916.16p0.parquet",
        table_name="nasa_table",
    )

    assert "nasa_table" in con.list_tables()
    assert t.op().name == "nasa_table"
    assert t.head().execute() is not None


def test_register_table(con):
    tab = pa.table({"x": [1, 2, 3]})
    con.create_table("my_table", tab)
    assert con.table("my_table").x.sum().execute() == 6


def test_register_pandas(con):
    df = pd.DataFrame({"x": [1, 2, 3]})
    con.create_table("my_pandas_table", df)
    assert con.table("my_pandas_table").x.sum().execute() == 6


def test_register_batches(con):
    batch = pa.record_batch([pa.array([1, 2, 3])], names=["x"])
    con.create_table("my_batch_table", batch)
    assert con.table("my_batch_table").x.sum().execute() == 6


def test_register_dataset(con):
    tab = pa.table({"x": [1, 2, 3]})
    dataset = ds.InMemoryDataset(tab)
    t = xo.memtable(dataset, name="my_table")
    assert con.execute(t.x.sum()) == 6


def test_register_memtable(con):
    data = pd.DataFrame({"a": [1, 2, 3, 4, 5], "b": [2, 3, 4, 5, 6]})
    t = xo.memtable(data)
    assert con.execute(t.a.sum()) == 15


@pytest.mark.gcs
def test_deferred_read_parquet_from_gcs(tmp_path):
    con = xo.connect()
    path = "gs://cloud-samples-data/bigquery/us-states/us-states.parquet"
    expr = (
        xo.deferred_read_parquet(path, con)
        .cache(
            cache=ParquetCache.from_kwargs(
                source=xo.duckdb.connect(), relative_path=tmp_path
            )
        )
        .limit(10)
    )

    assert not expr.execute().empty
    assert any(tmp_path.iterdir())


@pytest.mark.s3
def test_read_csv_from_s3_and_cache(tmp_path):
    con = xo.connect()

    # by providing the schema we avoid calling pandas
    schema = Schema.from_pyarrow(
        pa.schema(
            [
                pa.field("question", pa.string()),
                pa.field("product_description", pa.string()),
                pa.field("image_url", pa.string()),
                pa.field("label", pa.uint8(), nullable=True),
            ]
        )
    )

    path = "s3://humor-detection-pds/Humorous.csv"
    t = xo.deferred_read_csv(
        path,
        con,
        schema=schema,
        storage_options=FrozenDict(
            {
                "aws.region": "us-west-2",
            }
        ),
    )

    expr = t.cache(
        cache=ParquetCache.from_kwargs(
            source=xo.duckdb.connect(), relative_path=tmp_path
        )
    ).limit(10)

    assert not expr.execute().empty
    assert any(tmp_path.iterdir())


def test_get_object_metadata_local_filesystem(data_dir):
    url = data_dir / "parquet" / "functional_alltypes.parquet"
    metadata = xo.get_object_metadata(str(url.resolve()))

    assert isinstance(metadata, dict)


def test_get_object_metadata_https(ctx):
    from urllib.request import Request, urlopen

    url = "https://raw.githubusercontent.com/ibis-project/testing-data/refs/heads/master/csv/astronauts.csv"

    metadata = xo.get_object_metadata(url)
    assert isinstance(metadata, dict)

    request = Request(url, method="GET")
    with urlopen(request) as response:
        etag = response.headers.get("ETag")
        assert etag == metadata["e_tag"]


@pytest.mark.s3
def test_get_object_metadata_s3():
    metadata = xo.get_object_metadata(
        "s3://humor-detection-pds/Humorous.csv",
        storage_options={
            "aws.region": "us-west-2",
        },
    )

    assert isinstance(metadata, dict)


@pytest.mark.gcs
def test_get_object_metadata_gcs():
    metadata = xo.get_object_metadata(
        "gs://cloud-samples-data/bigquery/us-states/us-states.parquet",
    )
    assert isinstance(metadata, dict)


@pytest.mark.parametrize(
    "get_con",
    [
        lambda: xo.connect(),
        lambda: xo.datafusion.connect(),
        lambda: xo.duckdb.connect(),
        pytest.param(lambda: xo.postgres.connect_env(), marks=pytest.mark.postgres),
    ],
)
def test_expr_over_same_table_multiple_times(parquet_dir, get_con):
    ls_con = xo.connect()
    other_con = get_con()
    astronauts_path = parquet_dir.joinpath("astronauts.parquet")
    table_name = "astronauts"
    col = "id"

    astronauts_name = f"{ls_con.name}_{table_name}"
    astronauts = ls_con.register(astronauts_path, table_name=astronauts_name)

    if other_con.name == "postgres":
        t = other_con.table(table_name)
    else:
        t = other_con.read_parquet(astronauts_path, table_name=table_name)

    ls_table_name = f"{other_con.name}_{table_name}"
    ls_con.register(t, ls_table_name)
    other = ls_con.table(ls_table_name)

    expr = astronauts[[col]].distinct().join(other[[col]].distinct(), col)

    assert (first := expr.execute()) is not None
    assert (second := expr.execute()) is not None
    assert_frame_equal(first.sort_values(col), second.sort_values(col))


@pytest.mark.postgres
def test_read_postgres():
    import os

    uri = (
        f"postgres://{os.environ['POSTGRES_USER']}:"
        f"{os.environ['POSTGRES_PASSWORD']}@"
        f"{os.environ['POSTGRES_HOST']}:"
        f"{os.environ['POSTGRES_PORT']}/"
        f"{os.environ['POSTGRES_DATABASE']}"
    )
    t = xo.read_postgres(uri, table_name="astronauts")
    res = xo.execute(t)

    assert not res.empty



from abc import ABCMeta, abstractmethod
from typing import List

import pyarrow as pa
from xorq_datafusion._internal import (
    AggregateUDF,
    ContextProvider,
    DataFrame,
    LogicalPlan,
    LogicalPlanBuilder,
    OptimizerContext,
    OptimizerRule,
    ScalarUDF,
    SessionConfig,  # noqa: F401
    SessionContext,  # noqa: F401
    SessionState,  # noqa: F401
    Table,
    TableProvider,
    WindowUDF,
)


__all__ = [
    "SessionContext",
    "SessionState",
    "SessionConfig",
    "ContextProvider",
    "OptimizerContext",
    "LogicalPlan",
    "LogicalPlanBuilder",
    "ScalarUDF",
    "AggregateUDF",
    "OptimizerRule",
    "OptimizationRule",
    "TableProvider",
    "AbstractTableProvider",
    "Table",
    "Accumulator",
    "DataFrame",
    "WindowUDF",
    "WindowEvaluator",
]


class Accumulator(metaclass=ABCMeta):
    @abstractmethod
    def state(self) -> List[pa.Scalar]:
        pass

    @abstractmethod
    def update(self, values: pa.Array) -> None:
        pass

    @abstractmethod
    def merge(self, states: pa.Array) -> None:
        pass

    @abstractmethod
    def evaluate(self) -> pa.Scalar:
        pass


class OptimizationRule(metaclass=ABCMeta):
    @abstractmethod
    def try_optimize(self, plan: LogicalPlan) -> LogicalPlan:
        pass


class AbstractTableProvider(metaclass=ABCMeta):
    @abstractmethod
    def schema(self):
        pass

    @abstractmethod
    def scan(self, filters=None):
        pass


class WindowEvaluator(metaclass=ABCMeta):  # noqa: B024
    def memoize(self) -> None:  # noqa: B027
        pass

    def get_range(self, idx: int, num_rows: int) -> tuple[int, int]:
        return idx, idx + 1

    def is_causal(self) -> bool:
        return False

    def evaluate_all(self, values: list[pa.Array], num_rows: int) -> pa.Array:  # noqa: B027
        pass

    def evaluate(  # noqa: B027
        self, values: list[pa.Array], eval_range: tuple[int, int]
    ) -> pa.Scalar:
        pass

    def evaluate_all_with_rank(  # noqa: B027
        self, num_rows: int, ranks_in_partition: list[tuple[int, int]]
    ) -> pa.Array:
        pass

    def supports_bounded_execution(self) -> bool:
        return False

    def uses_window_frame(self) -> bool:
        return False

    def include_rank(self) -> bool:
        return False


def udf(func, input_types, return_type, volatility, name=None):
    """
    Create a new User Defined Function
    """
    if not callable(func):
        raise TypeError("`func` argument must be callable")
    if name is None:
        name = func.__qualname__.lower()
    return ScalarUDF(
        name=name,
        func=func,
        input_types=input_types,
        return_type=return_type,
        volatility=volatility,
    )


def udaf(accum, input_type, return_type, state_type, volatility, name=None):
    """
    Create a new User Defined Aggregate Function
    """
    if not issubclass(accum, Accumulator):
        raise TypeError("`accum` must implement the abstract base class Accumulator")
    if name is None:
        name = accum.__qualname__.lower()
    return AggregateUDF(
        name=name,
        accumulator=accum,
        input_type=input_type,
        return_type=return_type,
        state_type=state_type,
        volatility=volatility,
    )


def udwf(
    func: WindowEvaluator,
    input_types: pa.DataType | list[pa.DataType],
    return_type: pa.DataType,
    volatility: str,
    name: str | None = None,
) -> WindowUDF:
    """Create a new User-Defined Window Function.

    Args:
        func: The python function.
        input_types: The data types of the arguments to ``func``.
        return_type: The data type of the return value.
        volatility: See :py:class:`Volatility` for allowed values.
        name: A descriptive name for the function.

    Returns:
        A user-defined window function.
    """
    if not isinstance(func, WindowEvaluator):
        raise TypeError("`func` must implement the abstract base class WindowEvaluator")
    if name is None:
        name = func.__class__.__qualname__.lower()
    if isinstance(input_types, pa.DataType):
        input_types = [input_types]

    return WindowUDF(name, func, input_types, return_type, str(volatility))



from __future__ import annotations

from collections.abc import Mapping

import numpy as np
import pandas as pd
import pytest
from pytest import param

import xorq.api as xo
import xorq.expr.datatypes as dt
from xorq.tests.util import assert_frame_equal, assert_series_equal


@pytest.mark.parametrize(
    ("field", "expected"),
    [
        param(
            "a",
            [1.0, 2.0, 2.0, 3.0, 3.0, np.nan],
            id="a",
        ),
        param("b", ["apple", "banana", "banana", "orange", "orange", None], id="b"),
        param(
            "c",
            [2, 2, 3, 3, 4, np.nan],
            id="c",
        ),
    ],
)
def test_single_field(struct, field, expected):
    expr = struct.select(field=lambda t: t.abc[field]).order_by("field")
    result = expr.execute()
    assert_series_equal(
        result.field, pd.Series(expected, name="field"), check_dtype=False
    )


def test_all_fields(struct, struct_df):
    result = xo.execute(struct.abc)
    expected = struct_df.abc

    assert {
        row if not isinstance(row, Mapping) else tuple(row.items()) for row in result
    } == {
        row if not isinstance(row, Mapping) else tuple(row.items()) for row in expected
    }


_SIMPLE_DICT = {"a": 1, "b": "2", "c": 3.0}
_STRUCT_LITERAL = xo.struct(
    _SIMPLE_DICT,
    type="struct<a: int64, b: string, c: float64>",
)
_NULL_STRUCT_LITERAL = xo.null().cast("struct<a: int64, b: string, c: float64>")


@pytest.mark.parametrize("field", ["a", "b", "c"])
def test_literal(con, field):
    query = _STRUCT_LITERAL[field]
    dtype = query.type().to_pandas()
    result = pd.Series([con.execute(query)], dtype=dtype)
    result = result.replace({np.nan: None})
    expected = pd.Series([_SIMPLE_DICT[field]])
    assert_series_equal(result, expected.astype(dtype))


def test_struct_column(alltypes, alltypes_df):
    t = alltypes
    expr = t.select(s=xo.struct({"a": t.string_col, "b": 1, "c": t.bigint_col}))
    assert expr.s.type() == dt.Struct({"a": dt.string, "b": dt.int8, "c": dt.int64})
    result = xo.execute(expr)
    expected = pd.DataFrame(
        {
            "s": [
                {"a": a, "b": 1, "c": c}
                for a, c in zip(alltypes_df.string_col, alltypes_df.bigint_col)
            ]
        }
    )
    assert_frame_equal(result, expected)


def test_field_access_after_case(con):
    s = xo.struct({"a": 3})
    x = xo.cases((True, s), else_=xo.struct({"a": 4}))
    y = x.a
    assert con.to_pandas(y) == 3


def test_collect_into_struct(alltypes):
    from xorq.vendor.ibis import _

    t = alltypes
    expr = (
        t[_.string_col.isin(("0", "1"))]
        .group_by(group="string_col")
        .agg(
            val=lambda t: xo.struct(
                {"key": t.bigint_col.collect().cast("array<int64>")}
            )
        )
    )
    result = xo.execute(expr)
    assert result.shape == (2, 2)
    assert set(result.group) == {"0", "1"}
    val = result.val
    assert len(val.loc[result.group == "0"].iat[0]["key"]) == 730
    assert len(val.loc[result.group == "1"].iat[0]["key"]) == 730



from __future__ import annotations

import collections.abc
from abc import abstractmethod
from itertools import tee
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from public import public

from xorq.common.exceptions import ConflictingValuesError
from xorq.vendor.ibis.common.bases import Abstract, Hashable


if TYPE_CHECKING:
    from typing_extensions import Self

K = TypeVar("K", bound=collections.abc.Hashable)
V = TypeVar("V")


# The following classes provide an alternative to the `collections.abc` module
# which can be used with `ibis.common.bases` without metaclass conflicts but
# remains compatible with the `collections.abc` module. The main advantage is
# faster `isinstance` checks.


@collections.abc.Iterable.register
class Iterable(Abstract, Generic[V]):
    """Iterable abstract base class for quicker isinstance checks."""

    @abstractmethod
    def __iter__(self): ...


@collections.abc.Reversible.register
class Reversible(Iterable[V]):
    """Reverse iterable abstract base class for quicker isinstance checks."""

    @abstractmethod
    def __reversed__(self): ...


@collections.abc.Iterator.register
class Iterator(Iterable[V]):
    """Iterator abstract base class for quicker isinstance checks."""

    @abstractmethod
    def __next__(self): ...

    def __iter__(self):
        return self


@collections.abc.Sized.register
class Sized(Abstract):
    """Sized abstract base class for quicker isinstance checks."""

    @abstractmethod
    def __len__(self): ...


@collections.abc.Container.register
class Container(Abstract, Generic[V]):
    """Container abstract base class for quicker isinstance checks."""

    @abstractmethod
    def __contains__(self, x): ...


@collections.abc.Collection.register
class Collection(Sized, Iterable[V], Container[V]):
    """Collection abstract base class for quicker isinstance checks."""


@collections.abc.Sequence.register
class Sequence(Reversible[V], Collection[V]):
    """Sequence abstract base class for quicker isinstance checks."""

    @abstractmethod
    def __getitem__(self, index): ...

    def __iter__(self):
        i = 0
        try:
            while True:
                yield self[i]
                i += 1
        except IndexError:
            return

    def __contains__(self, value):
        return any(v is value or v == value for v in self)

    def __reversed__(self):
        for i in reversed(range(len(self))):
            yield self[i]

    def index(self, value, start=0, stop=None):
        if start is not None and start < 0:
            start = max(len(self) + start, 0)
        if stop is not None and stop < 0:
            stop += len(self)

        i = start
        while stop is None or i < stop:
            try:
                v = self[i]
            except IndexError:
                break
            if v is value or v == value:
                return i
            i += 1
        raise ValueError

    def count(self, value):
        return sum(1 for v in self if v is value or v == value)


@collections.abc.Mapping.register
class Mapping(Collection[K], Generic[K, V]):
    """Mapping abstract base class for quicker isinstance checks."""

    @abstractmethod
    def __getitem__(self, key): ...

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __contains__(self, key):
        try:
            self[key]
        except KeyError:
            return False
        else:
            return True

    def keys(self):
        return collections.abc.KeysView(self)

    def items(self):
        return collections.abc.ItemsView(self)

    def values(self):
        return collections.abc.ValuesView(self)

    def __eq__(self, other):
        if not isinstance(other, collections.abc.Mapping):
            return NotImplemented
        return dict(self.items()) == dict(other.items())


@public
class MapSet(Mapping[K, V]):
    """A mapping that also supports set-like operations.

    It is an altered version of `collections.abc.Mapping` that supports set-like
    operations. The `__iter__`, `__len__`, and `__getitem__` methods must be
    implemented.

    The set-like operations' other operand must be a `Mapping`. If the two
    operands contain common keys but with different values, then the operation
    becomes ambiguous and an exception will be raised.

    Examples
    --------
    >>> from xorq.vendor.ibis.common.collections import MapSet
    >>> class MyMap(MapSet):
    ...     __slots__ = ("_data",)
    ...
    ...     def __init__(self, *args, **kwargs):
    ...         self._data = dict(*args, **kwargs)
    ...
    ...     def __iter__(self):
    ...         return iter(self._data)
    ...
    ...     def __len__(self):
    ...         return len(self._data)
    ...
    ...     def __getitem__(self, key):
    ...         return self._data[key]
    ...
    ...     def __repr__(self):
    ...         return f"MyMap({repr(self._data)})"
    >>> m = MyMap(a=1, b=2)
    >>> n = dict(a=1, b=2, c=3)
    >>> m <= n
    True
    >>> m < n
    True
    >>> n - m
    MyMap({'c': 3})
    >>> m & n
    MyMap({'a': 1, 'b': 2})
    >>> m | n
    MyMap({'a': 1, 'b': 2, 'c': 3})

    """

    def _check_conflict(self, other: collections.abc.Mapping) -> set[K]:
        # Check if there are conflicting key-value pairs between self and other.
        # A key-value pair is conflicting if the key is the same but the value is
        # different.
        common_keys = self.keys() & other.keys()
        conflicts = {
            (key, self[key], other[key])
            for key in common_keys
            if self[key] != other[key]
        }
        if conflicts:
            raise ConflictingValuesError(conflicts)
        return common_keys

    def __ge__(self, other: collections.abc.Mapping) -> bool:
        if not isinstance(other, collections.abc.Mapping):
            return NotImplemented
        common_keys = self._check_conflict(other)
        return other.keys() == common_keys

    def __gt__(self, other: collections.abc.Mapping) -> bool:
        if not isinstance(other, collections.abc.Mapping):
            return NotImplemented
        return len(self) > len(other) and self.__ge__(other)

    def __le__(self, other: collections.abc.Mapping) -> bool:
        if not isinstance(other, collections.abc.Mapping):
            return NotImplemented
        common_keys = self._check_conflict(other)
        return self.keys() == common_keys

    def __lt__(self, other: collections.abc.Mapping) -> bool:
        if not isinstance(other, collections.abc.Mapping):
            return NotImplemented
        return len(self) < len(other) and self.__le__(other)

    def __and__(self, other: collections.abc.Mapping) -> Self:
        if not isinstance(other, collections.abc.Mapping):
            return NotImplemented
        common_keys = self._check_conflict(other)
        intersection = {k: v for k, v in self.items() if k in common_keys}
        return self.__class__(intersection)

    def __sub__(self, other: collections.abc.Mapping) -> Self:
        if not isinstance(other, collections.abc.Mapping):
            return NotImplemented
        common_keys = self._check_conflict(other)
        difference = {k: v for k, v in self.items() if k not in common_keys}
        return self.__class__(difference)

    def __rsub__(self, other: collections.abc.Mapping) -> Self:
        if not isinstance(other, collections.abc.Mapping):
            return NotImplemented
        common_keys = self._check_conflict(other)
        difference = {k: v for k, v in other.items() if k not in common_keys}
        return self.__class__(difference)

    def __or__(self, other: collections.abc.Mapping) -> Self:
        if not isinstance(other, collections.abc.Mapping):
            return NotImplemented
        self._check_conflict(other)
        union = {**self, **other}
        return self.__class__(union)

    def __xor__(self, other: collections.abc.Mapping) -> Self:
        if not isinstance(other, collections.abc.Mapping):
            return NotImplemented
        left = self - other
        right = other - self
        left._check_conflict(right)
        union = {**left, **right}
        return self.__class__(union)

    def isdisjoint(self, other: collections.abc.Mapping) -> bool:
        common_keys = self._check_conflict(other)
        return not common_keys


@public
class FrozenDict(dict, Mapping[K, V], Hashable):
    __slots__ = ("__precomputed_hash__",)
    __precomputed_hash__: int

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        hashable = frozenset(self.items())
        object.__setattr__(self, "__precomputed_hash__", hash(hashable))

    def __hash__(self) -> int:
        return self.__precomputed_hash__

    def __setitem__(self, key: K, value: V) -> None:
        raise TypeError(
            f"'{self.__class__.__name__}' object does not support item assignment"
        )

    def __setattr__(self, name: str, _: Any) -> None:
        raise TypeError(f"Attribute {name!r} cannot be assigned to frozendict")

    def __reduce__(self) -> tuple:
        return (self.__class__, (dict(self),))


@public
class FrozenOrderedDict(FrozenDict[K, V]):
    def __init__(self, *args, **kwargs):
        super(FrozenDict, self).__init__(*args, **kwargs)
        hashable = tuple(self.items())
        object.__setattr__(self, "__precomputed_hash__", hash(hashable))

    def __hash__(self) -> int:
        return self.__precomputed_hash__

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, collections.abc.Mapping):
            return NotImplemented
        return tuple(self.items()) == tuple(other.items())

    def __ne__(self, other: Any) -> bool:
        return not self == other


class RewindableIterator(Iterator[V]):
    """Iterator that can be rewound to a checkpoint.

    Examples
    --------
    >>> it = RewindableIterator(range(5))
    >>> next(it)
    0
    >>> next(it)
    1
    >>> it.checkpoint()
    >>> next(it)
    2
    >>> next(it)
    3
    >>> it.rewind()
    >>> next(it)
    2
    >>> next(it)
    3
    >>> next(it)
    4

    """

    __slots__ = ("_iterator", "_checkpoint")

    def __init__(self, iterable):
        self._iterator = iter(iterable)
        self._checkpoint = None

    def __next__(self):
        return next(self._iterator)

    def rewind(self):
        """Rewind the iterator to the last checkpoint."""
        if self._checkpoint is None:
            raise ValueError("No checkpoint to rewind to.")
        self._iterator, self._checkpoint = tee(self._checkpoint)

    def checkpoint(self):
        """Create a checkpoint of the current iterator state."""
        self._iterator, self._checkpoint = tee(self._iterator)


# Need to provide type hint as else a static type checker does not recognize
# that frozendict exists in this module
frozendict: type[FrozenDict]
public(frozendict=FrozenDict)



import pathlib
import runpy

import pytest
from pytest import param

import xorq.api as xo


KEY_PREFIX = xo.config.options.cache.key_prefix
LIBRARY_SCRIPTS = (
    "pandas_example",
    "flight_dummy_exchanger",
)
GCS_SCRIPTS = ("gcstorage_example",)
NON_TESTABLE = (
    "mcp_flight_server.py",
    "duckdb_flight_example.py",
    "complex_cached_expr.py",
    "xorq_build_and_run.py",
    "weather_flight.py",
    "gizmosql_demo.py",
)

file_path = pathlib.Path(__file__).absolute()
root = file_path.parent
examples_dir = file_path.parents[3] / "examples"
scripts = (p for p in examples_dir.glob("*.py") if p.name not in NON_TESTABLE)


def teardown_function():
    """Remove any generated parquet cache files"""
    for path in root.glob(f"{KEY_PREFIX}*"):
        path.unlink(missing_ok=True)

    for path in pathlib.Path.cwd().glob(f"{KEY_PREFIX}*"):
        path.unlink(missing_ok=True)


def maybe_library(name: str):
    return pytest.mark.library if name in LIBRARY_SCRIPTS else ()


def maybe_gcs(name: str):
    return pytest.mark.gcs if name in GCS_SCRIPTS else ()


def maybe_marks(name: str):
    fs = (maybe_library, maybe_gcs)
    return tuple(filter(None, (f(name) for f in fs)))


@pytest.mark.parametrize(
    "script",
    [
        param(script, id=script.stem, marks=maybe_marks(script.stem))
        for script in scripts
    ],
)
def test_script_execution(script):
    dct = runpy.run_path(str(script), run_name="__pytest_main__")
    assert dct.get("pytest_examples_passed")



from __future__ import annotations

from xorq.vendor.ibis.expr.datatypes.cast import *  # noqa: F403
from xorq.vendor.ibis.expr.datatypes.core import *  # noqa: F403
from xorq.vendor.ibis.expr.datatypes.parse import *  # noqa: F403
from xorq.vendor.ibis.expr.datatypes.value import *  # noqa: F403


halffloat = float16  # noqa: F405
float = float64  # noqa: F405
double = float64  # noqa: F405
int = int64  # noqa: F405
uint_ = uint64  # noqa: F405
bool = boolean  # noqa: F405
str = string  # noqa: F405
bytes = binary  # noqa: F405

validate_type = dtype  # noqa: F405



import functools
import time
from typing import Optional

import pyarrow as pa
from opentelemetry import metrics
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import (
    MetricExporter,
    MetricExportResult,
    PeriodicExportingMetricReader,
)

from xorq.common.utils.logging_utils import get_print_logger


__all__ = [
    "setup_console_metrics",
    "instrument_rpc",
    "instrument_reader",
    "InstrumentedWriter",
]

logger = get_print_logger()
# placeholders for instruments; initialized in setup_console_metrics()
_request_counter = None
_duration_hist = None
_streams_counter = None
_bytes_counter = None
_rows_counter = None
_throughput_hist = None


class SimpleConsoleMetricExporter(MetricExporter):
    def export(self, metrics_data, timeout_millis: int = 0, **kwargs):
        for rm in getattr(metrics_data, "resource_metrics", []):
            for scope in getattr(rm, "scope_metrics", []):
                for metric in getattr(scope, "metrics", []):
                    name = metric.name
                    data = metric.data
                    for pt in data.data_points:
                        if hasattr(pt, "value"):
                            val = pt.value
                            attrs = pt.attributes or {}
                            suffix = (
                                "{"
                                + ",".join(f"{k}={v}" for k, v in attrs.items())
                                + "}"
                                if attrs
                                else ""
                            )
                            logger.info(f"{name}{suffix} {val}")
                        else:
                            try:
                                for q, v in pt.value.quantiles:
                                    logger.info(f"{name}{{quantile={q}}} {v}")
                            except Exception:
                                pass
        return MetricExportResult.SUCCESS

    def shutdown(self, *args, **kwargs) -> None:
        return None

    def force_flush(self, *args, **kwargs) -> bool:
        return True


@functools.cache
def setup_console_metrics(
    interval_ms: int = 2000,
    meter_name: str = "xorq.flight_server",
    prometheus_port: Optional[int] = None,
):
    exporter = SimpleConsoleMetricExporter()

    readers = [
        PeriodicExportingMetricReader(exporter, export_interval_millis=interval_ms)
    ]
    # optional Prometheus endpoint
    if prometheus_port is not None:
        try:
            import prometheus_client
            from opentelemetry.exporter.prometheus import PrometheusMetricReader
        except ImportError:
            logger.warning(
                "Prometheus support requires 'opentelemetry-exporter-prometheus' and 'prometheus-client'"
            )
        else:
            prom_reader = PrometheusMetricReader()
            readers.append(prom_reader)
            prometheus_client.start_http_server(prometheus_port)
            logger.info(f"Prometheus metrics available at :{prometheus_port}")
    # include resource attributes for backend grouping
    import socket

    from opentelemetry.sdk.resources import Resource

    resource = Resource.create(
        {
            "service.name": "xorq-flight",
            "service.instance.id": socket.gethostname(),
        }
    )
    provider = MeterProvider(resource=resource, metric_readers=readers)
    metrics.set_meter_provider(provider)
    meter = metrics.get_meter(meter_name)
    logger.info(f"console metrics enabled, interval={interval_ms} ms")
    global \
        _request_counter, \
        _duration_hist, \
        _streams_counter, \
        _bytes_counter, \
        _rows_counter, \
        _throughput_hist
    _request_counter = meter.create_counter(
        "flight_server.requests_total", description="Total Flight RPC requests"
    )
    _duration_hist = meter.create_histogram(
        "flight_server.request_duration_seconds", unit="s"
    )
    _streams_counter = meter.create_up_down_counter("flight_server.active_streams")
    _bytes_counter = meter.create_counter("flight_server.bytes_total", unit="By")
    _rows_counter = meter.create_counter("flight_server.rows_total")
    _throughput_hist = meter.create_histogram(
        "flight_server.throughput_rows_per_sec", unit="1/s"
    )
    return meter


class _Recorder:
    """Context-manager that tracks one RPC end-to-end."""

    def __init__(self, method: str):
        self.method = method
        self.started = time.time()

        if _request_counter is not None and _streams_counter is not None:
            _request_counter.add(1, {"method": method})
            _streams_counter.add(1)
        self._bytes = 0
        self._rows = 0

    def record_batch(self, batch: pa.RecordBatch, direction: str):
        if not hasattr(batch, "get_total_buffer_size") and hasattr(batch, "data"):
            rb = batch.data
        else:
            rb = batch
        size = rb.get_total_buffer_size()
        rows = rb.num_rows
        labels = {"method": self.method, "dir": direction}

        if _bytes_counter is not None and _streams_counter is not None:
            _bytes_counter.add(size, labels)
            _rows_counter.add(rows, labels)
        self._bytes += size
        self._rows += rows

    def finish(self):
        dur = time.time() - self.started

        if (
            _duration_hist is not None
            and _throughput_hist is not None
            and _streams_counter is not None
        ):
            _duration_hist.record(dur, {"method": self.method})
            if dur > 0 and self._rows:
                thr = self._rows / dur
                _throughput_hist.record(thr, {"method": self.method})
            _streams_counter.add(-1)


class InstrumentedReader:
    def __init__(self, reader, rec: _Recorder, direction: str = "out"):
        self.reader = reader
        self.rec = rec
        self.direction = direction

    @property
    def schema(self):
        return self.reader.schema

    def __iter__(self):
        for batch in self.reader:
            if not hasattr(batch, "get_total_buffer_size") and hasattr(batch, "data"):
                rb = batch.data
            else:
                rb = batch
            self.rec.record_batch(rb, self.direction)
            yield batch


def instrument_reader(reader: pa.RecordBatchReader, rec: _Recorder, *, direction="out"):
    def gen():
        for batch in reader:
            if not hasattr(batch, "get_total_buffer_size") and hasattr(batch, "data"):
                rb = batch.data
            else:
                rb = batch
            rec.record_batch(rb, direction)
            yield rb

    return pa.RecordBatchReader.from_batches(reader.schema, gen())


class InstrumentedWriter:
    def __init__(self, raw_writer, rec: _Recorder, *, direction="out"):
        self._w = raw_writer
        self._rec = rec
        self._dir = direction

    def begin(self, *args, **kwargs):
        return self._w.begin(*args, **kwargs)

    def write_metadata(self, metadata):
        return self._w.write_metadata(metadata)

    def write_batch(self, batch):
        self._rec.record_batch(batch, self._dir)
        return self._w.write_batch(batch)

    def write_with_metadata(self, batch, metadata):
        if batch:
            self._rec.record_batch(batch, self._dir)
        return self._w.write_with_metadata(batch, metadata)


def instrument_rpc(method_name):
    def decorator(func):
        def wrapper(self, context, *args, **kwargs):
            rec = _Recorder(method_name)
            try:
                return func(self, context, rec, *args, **kwargs)
            finally:
                rec.finish()

        return wrapper

    return decorator



import xorq.vendor.ibis as ibis
from xorq.ibis_yaml.tests.conftest import get_dtype_yaml


def test_explicit_cast(compiler):
    expr = ibis.literal(42).cast("float64")
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)

    assert expression["op"] == "Cast"
    assert expression["arg"]["op"] == "Literal"
    assert expression["arg"]["value"] == 42
    assert dtype_yaml["type"] == "Float64"

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_implicit_cast(compiler):
    i = ibis.literal(1)
    f = ibis.literal(2.5)
    expr = i + f
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)

    assert expression["op"] == "Add"
    assert get_dtype_yaml(yaml_dict, expression["left"])["type"] == "Int8"
    assert get_dtype_yaml(yaml_dict, expression["right"])["type"] == "Float64"
    assert dtype_yaml["type"] == "Float64"

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_string_cast(compiler):
    expr = ibis.literal("42").cast("int64")
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)

    assert expression["op"] == "Cast"
    assert expression["arg"]["value"] == "42"
    assert dtype_yaml["type"] == "Int64"

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_timestamp_cast(compiler):
    expr = ibis.literal("2024-01-01").cast("timestamp")
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)

    assert expression["op"] == "Cast"
    assert expression["arg"]["value"] == "2024-01-01"
    assert dtype_yaml["type"] == "Timestamp"

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)



import functools
import itertools
import os

import _pytest
import pytest

import xorq.api as xo
from xorq.backends import _get_backend_names
from xorq.caching import ParquetSnapshotCache, SourceSnapshotCache
from xorq.vendor import ibis


snowflake_credentials_varnames = (
    "SNOWFLAKE_PRIVATE_KEY",
    "SNOWFLAKE_PRIVATE_KEY_PWD",
    "SNOWFLAKE_USER",
)
have_snowflake_credentials = all(
    os.environ.get(varname) for varname in snowflake_credentials_varnames
)

KEY_PREFIX = xo.config.options.cache.key_prefix


def _get_backend_from_parts(parts: tuple[str, ...]) -> str | None:
    """Return the backend part of a test file's path parts.

    Examples
    --------
    >>> _get_backend_from_parts(("/", "ibis", "backends", "sqlite", "tests"))
    "sqlite"
    """
    try:
        index = parts.index("backends")
    except ValueError:
        return None
    else:
        return parts[index + 1]


def pytest_ignore_collect(collection_path, config):
    # get the backend path part
    backend = _get_backend_from_parts(collection_path.parts)
    if backend is None or backend not in _get_backend_names():
        return False

    # we evaluate the marker early so that we don't trigger
    # an import of conftest files for the backend, which will
    # import the backend and thus require dependencies that may not
    # exist
    #
    # alternatives include littering library code with pytest.importorskips
    # and moving all imports close to their use site
    #
    # the latter isn't tenable for backends that use multiple dispatch
    # since the rules are executed at import time
    mark_expr = config.getoption("-m")
    # we can't let the empty string pass through, since `'' in s` is `True` for
    # any `s`; if no marker was passed don't ignore the collection of `path`
    if not mark_expr:
        return False
    expr = _pytest.mark.expression.Expression.compile(mark_expr)
    # we check the "backend" marker as well since if that's passed
    # any file matching a backend should be skipped
    keep = expr.evaluate(lambda s, **_kw: s in (backend, "backend"))
    return not keep


def pytest_collection_modifyitems(session, config, items):
    all_backends = _get_backend_names()
    additional_markers = []

    unrecognized_backends = set()
    for item in items:
        # add the backend marker to any tests are inside "xorq/backends"
        parts = item.path.parts
        backend = _get_backend_from_parts(parts)
        if backend is not None and backend in all_backends:
            item.add_marker(getattr(pytest.mark, backend))
            item.add_marker(pytest.mark.backend)
        elif "backends" not in parts and not tuple(
            itertools.chain(
                *(item.iter_markers(name=name) for name in all_backends),
                item.iter_markers(name="backend"),
            )
        ):
            # anything else is a "core" test and is run by default
            if not any(item.iter_markers(name="benchmark")):
                item.add_marker(pytest.mark.core)

    if unrecognized_backends:
        raise pytest.PytestCollectionWarning("\n" + "\n".join(unrecognized_backends))

    for item, markers in additional_markers:
        for marker in markers:
            item.add_marker(marker)


@functools.cache
def maybe_snowflake_decrypt_failure():
    try:
        SKU = pytest.importorskip("xorq.common.utils.snowflake_keypair_utils")
        kwargs = {
            varname[len("SNOWFLAKE_") :].lower(): os.environ.get(varname)
            for varname in snowflake_credentials_varnames
        }
        SKU.maybe_decrypt_private_key(kwargs)
    except Exception as e:
        return e


def pytest_runtest_setup(item):
    if any(mark.name == "snowflake" for mark in item.iter_markers()):
        pytest.importorskip("snowflake.connector")
        if not have_snowflake_credentials:
            pytest.skip("cannot run snowflake tests without snowflake creds")
        if e := maybe_snowflake_decrypt_failure():
            pytest.fail(f"cannot decrpyt snowflake creds '{e}'")


def get_cache_uncached(expr):
    assert expr.ls.is_cached
    cache = expr.ls.cache
    uncached = expr.ls.uncached_one
    return (cache, uncached)


@pytest.fixture(scope="session")
def xo_con(parquet_dir):
    conn = xo.connect()

    for name in ("astronauts", "functional_alltypes", "awards_players", "batting"):
        conn.read_parquet(parquet_dir / f"{name}.parquet", name)

    return conn


@pytest.fixture(scope="session")
def alltypes(xo_con):
    return xo_con.table("functional_alltypes")


@pytest.fixture(scope="session")
def alltypes_df(alltypes):
    return alltypes.execute()


@pytest.fixture(scope="session")
def awards_players(xo_con):
    return xo_con.table("awards_players")


@pytest.fixture(scope="session")
def batting(xo_con):
    return xo_con.table("batting")


@pytest.fixture(scope="session")
def con_snapshot(xo_con, alltypes_df):
    def _con_snapshot(_alltypes_df, _xo_con, df_con):
        group_by = "year"
        name = ibis.util.gen_name("tmp_table")
        # create a temp table we can mutate
        table = df_con.create_table(name, _alltypes_df)
        cached_expr = (
            table.group_by(group_by)
            .agg({f"count_{col}": table[col].count() for col in table.columns})
            .cache(cache=SourceSnapshotCache.from_kwargs(source=_xo_con))
        )
        (cache, uncached) = get_cache_uncached(cached_expr)
        # test preconditions
        assert not cache.exists(uncached)
        # test cache creation
        executed0 = cached_expr.execute()
        assert cache.exists(uncached)
        # test cache use
        executed1 = cached_expr.execute()
        assert executed0.equals(executed1)
        # test NO cache invalidation
        df_con.insert(name, _alltypes_df)
        executed2 = cached_expr.execute()
        executed3 = cached_expr.ls.uncached.execute()
        assert executed0.equals(executed2)
        assert not executed0.equals(executed3)
        # key does not have key prefix
        assert cache.calc_key(uncached).count(KEY_PREFIX) == 1

    return functools.partial(_con_snapshot, alltypes_df, xo_con)


@pytest.fixture(scope="session")
def con_cross_source_snapshot(xo_con, alltypes_df):
    def _con_cross_source_snapshot(_alltypes_df, _con, expr_con):
        group_by = "year"
        name = ibis.util.gen_name("tmp_table")
        # create a temp table we can mutate
        table = expr_con.create_table(name, _alltypes_df)
        cache = ParquetSnapshotCache.from_kwargs(source=_con)
        expr = table.group_by(group_by).agg(
            {f"count_{col}": table[col].count() for col in table.columns}
        )
        cached_expr = expr.cache(cache=cache)
        # test preconditions
        assert not cache.exists(expr)  # the expr is not cached
        assert cache.storage.source is not expr_con  # the cache is cross source
        # test cache creation
        df = cached_expr.execute()
        assert not df.empty
        assert cache.exists(expr)
        # test cache use
        executed1 = cached_expr.execute()
        assert df.equals(executed1)
        # test NO cache invalidation
        expr_con.insert(name, _alltypes_df)
        executed2 = cached_expr.execute()
        executed3 = cached_expr.ls.uncached.execute()
        assert df.equals(executed2)
        assert not df.equals(executed3)

    return functools.partial(_con_cross_source_snapshot, alltypes_df, xo_con)


@pytest.fixture(scope="session")
def con_cache_find_backend(parquet_dir):
    def _con_cache_find_backend(_parquet_dir, cls, conn):
        astronauts_path = _parquet_dir / "astronauts.parquet"
        cache = cls.from_kwargs(source=conn)
        expr = conn.read_parquet(astronauts_path).cache(cache=cache)
        assert expr._find_backend()._profile == conn._profile

    return functools.partial(_con_cache_find_backend, parquet_dir)



import ibis

from xorq.common.utils.ibis_utils import from_ibis


def test_explicit_cast():
    expr = ibis.literal(42).cast("float64")

    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None


def test_implicit_cast():
    i = ibis.literal(1)
    f = ibis.literal(2.5)
    expr = i + f

    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None


def test_string_cast():
    expr = ibis.literal("42").cast("int64")

    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None


def test_timestamp_cast():
    expr = ibis.literal("2024-01-01").cast("timestamp")

    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None



# https://github.com/cloudpipe/cloudpickle/issues/178#issuecomment-975735397
import copyreg
import functools


lru_cache_type = type(functools.lru_cache()(lambda: None))


def new_lru_cache(func, cache_kwargs):
    return functools.lru_cache(**cache_kwargs)(func)


def _pickle_lru_cache(obj: lru_cache_type):
    make_params = getattr(obj, "cache_parameters", dict)
    return new_lru_cache, (obj.__wrapped__, make_params())


copyreg.pickle(lru_cache_type, _pickle_lru_cache)



from __future__ import annotations

import numpy as np
import pandas as pd
import pyarrow as pa
import pytest
from pytest import param

import xorq.api as xo
import xorq.vendor.ibis.expr.operations as ops
from xorq.tests.util import assert_frame_equal


@pytest.fixture
def client():
    return xo.pandas.connect(
        {
            "df": pd.DataFrame({"a": [1, 2, 3], "b": list("abc")}),
            "df_unknown": pd.DataFrame({"array_of_strings": [["a", "b"], [], ["c"]]}),
        }
    )


@pytest.fixture
def table(client):
    return client.table("df")


@pytest.fixture
def test_data():
    return pd.DataFrame({"A": [1, 2, 3, 4, 5], "B": list("abcde")})


def test_connect_no_args():
    con = xo.pandas.connect()
    assert dict(con.tables) == {}


def test_client_table(table):
    assert isinstance(table.op(), ops.DatabaseTable)


@pytest.mark.parametrize(
    "lamduh",
    [(lambda df: df), (lambda df: pa.Table.from_pandas(df))],
    ids=["dataframe", "pyarrow table"],
)
def test_create_table(client, test_data, lamduh):
    test_data = lamduh(test_data)
    client.create_table("testing", obj=test_data)
    assert "testing" in client.list_tables()
    client.create_table("testingschema", schema=client.get_schema("testing"))
    assert "testingschema" in client.list_tables()


def test_literal(client):
    lit = xo.literal(1)
    result = client.execute(lit)
    assert result == 1


def test_list_tables(client):
    assert client.list_tables(like="df_unknown")
    assert not client.list_tables(like="not_in_the_database")
    assert client.list_tables()


def test_drop(table):
    table = table.mutate(c=table.a)
    expr = table.drop("a")
    result = expr.execute()
    expected = table[["b", "c"]].execute()
    assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    "unit",
    [
        "Y",
        "M",
        "D",
        "h",
        "m",
        "s",
        "ms",
        "us",
        "ns",
        param("ps", marks=pytest.mark.xfail),
        param("fs", marks=pytest.mark.xfail),
        param("as", marks=pytest.mark.xfail),
    ],
)
def test_datetime64_infer(client, unit):
    value = np.datetime64("2018-01-02", unit)
    expr = xo.literal(value, type="timestamp")
    result = client.execute(expr)
    assert result == pd.Timestamp(value).to_pydatetime()



# Copyright 2014 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""User API for grouping operations."""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated

from public import public

import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
import xorq.vendor.ibis.expr.types as ir
from xorq.vendor import ibis
from xorq.vendor.ibis.common.grounds import Concrete
from xorq.vendor.ibis.common.patterns import Length  # noqa: TCH001
from xorq.vendor.ibis.common.typing import VarTuple  # noqa: TCH001
from xorq.vendor.ibis.expr.rewrites import rewrite_window_input


if TYPE_CHECKING:
    from collections.abc import Sequence


@public
class GroupedTable(Concrete):
    """An intermediate table expression to hold grouping information."""

    table: ops.Relation
    groupings: Annotated[VarTuple[ops.Value], Length(at_least=1)]
    orderings: VarTuple[ops.SortKey] = ()
    havings: VarTuple[ops.Value[dt.Boolean]] = ()

    def __getitem__(self, args):
        # Shortcut for projection with window functions
        return self.select(*args)

    def __getattr__(self, attr):
        try:
            field = getattr(self.table.to_expr(), attr)
        except AttributeError as e:
            raise AttributeError(f"GroupedTable has no attribute {attr}") from e

        if isinstance(field, ir.NumericValue):
            return GroupedNumbers(field, self)
        else:
            return GroupedArray(field, self)

    def aggregate(self, *metrics, **kwds) -> ir.Table:
        """Compute aggregates over a group by."""
        metrics = self.table.to_expr().bind(*metrics, **kwds)
        return self.table.to_expr().aggregate(
            metrics, by=self.groupings, having=self.havings
        )

    agg = aggregate

    def having(self, *predicates: ir.BooleanScalar) -> GroupedTable:
        """Add a post-aggregation result filter `expr`.

        ::: {.callout-warning}
        ## Expressions like `x is None` return `bool` and **will not** generate a SQL comparison to `NULL`
        :::

        Parameters
        ----------
        predicates
            Expressions that filters based on an aggregate value.

        Returns
        -------
        GroupedTable
            A grouped table expression
        """
        table = self.table.to_expr()
        havings = table.bind(*predicates)
        return self.copy(havings=self.havings + havings)

    def order_by(self, *by: ir.Value) -> GroupedTable:
        """Sort a grouped table expression by `expr`.

        Notes
        -----
        This API call is ignored in aggregations.

        Parameters
        ----------
        by
            Expressions to order the results by

        Returns
        -------
        GroupedTable
            A sorted grouped GroupedTable
        """
        table = self.table.to_expr()
        orderings = table.bind(*by)
        return self.copy(orderings=self.orderings + orderings)

    def mutate(
        self, *exprs: ir.Value | Sequence[ir.Value], **kwexprs: ir.Value
    ) -> ir.Table:
        """Return a table projection with window functions applied.

        Any arguments can be functions.

        Parameters
        ----------
        exprs
            List of expressions
        kwexprs
            Expressions

        Examples
        --------
        >>> import ibis
        >>> import ibis.selectors as s
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch()
        >>> t
        ┏━━━━━━━━━┳━━━━━━━━━━━┳━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━┳━━━┓
        ┃ species ┃ island    ┃ bill_length_mm ┃ bill_depth_mm ┃ flipper_length_mm ┃ … ┃
        ┡━━━━━━━━━╇━━━━━━━━━━━╇━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━╇━━━┩
        │ string  │ string    │ float64        │ float64       │ int64             │ … │
        ├─────────┼───────────┼────────────────┼───────────────┼───────────────────┼───┤
        │ Adelie  │ Torgersen │           39.1 │          18.7 │               181 │ … │
        │ Adelie  │ Torgersen │           39.5 │          17.4 │               186 │ … │
        │ Adelie  │ Torgersen │           40.3 │          18.0 │               195 │ … │
        │ Adelie  │ Torgersen │           NULL │          NULL │              NULL │ … │
        │ Adelie  │ Torgersen │           36.7 │          19.3 │               193 │ … │
        │ Adelie  │ Torgersen │           39.3 │          20.6 │               190 │ … │
        │ Adelie  │ Torgersen │           38.9 │          17.8 │               181 │ … │
        │ Adelie  │ Torgersen │           39.2 │          19.6 │               195 │ … │
        │ Adelie  │ Torgersen │           34.1 │          18.1 │               193 │ … │
        │ Adelie  │ Torgersen │           42.0 │          20.2 │               190 │ … │
        │ …       │ …         │              … │             … │                 … │ … │
        └─────────┴───────────┴────────────────┴───────────────┴───────────────────┴───┘
        >>> (
        ...     t.select("species", "bill_length_mm")
        ...     .group_by("species")
        ...     .mutate(centered_bill_len=ibis._.bill_length_mm - ibis._.bill_length_mm.mean())
        ...     .order_by(s.all())
        ... )
        ┏━━━━━━━━━┳━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━┓
        ┃ species ┃ bill_length_mm ┃ centered_bill_len ┃
        ┡━━━━━━━━━╇━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━┩
        │ string  │ float64        │ float64           │
        ├─────────┼────────────────┼───────────────────┤
        │ Adelie  │           32.1 │         -6.691391 │
        │ Adelie  │           33.1 │         -5.691391 │
        │ Adelie  │           33.5 │         -5.291391 │
        │ Adelie  │           34.0 │         -4.791391 │
        │ Adelie  │           34.1 │         -4.691391 │
        │ Adelie  │           34.4 │         -4.391391 │
        │ Adelie  │           34.5 │         -4.291391 │
        │ Adelie  │           34.6 │         -4.191391 │
        │ Adelie  │           34.6 │         -4.191391 │
        │ Adelie  │           35.0 │         -3.791391 │
        │ …       │              … │                 … │
        └─────────┴────────────────┴───────────────────┘

        Returns
        -------
        Table
            A table expression with window functions applied
        """
        exprs = self._selectables(*exprs, **kwexprs)
        return self.table.to_expr().mutate(exprs)

    def select(self, *exprs, **kwexprs) -> ir.Table:
        """Project new columns out of the grouped table.

        See Also
        --------
        [`GroupedTable.mutate`](#ibis.expr.types.groupby.GroupedTable.mutate)
        """
        exprs = self._selectables(*exprs, **kwexprs)
        return self.table.to_expr().select(exprs)

    def _selectables(self, *exprs, **kwexprs):
        """Project new columns out of the grouped table.

        See Also
        --------
        [`GroupedTable.mutate`](#ibis.expr.types.groupby.GroupedTable.mutate)
        """
        table = self.table.to_expr()
        values = table.bind(*exprs, **kwexprs)
        window = ibis.window(group_by=self.groupings, order_by=self.orderings)
        return [rewrite_window_input(expr.op(), window).to_expr() for expr in values]

    projection = select

    def over(
        self,
        window=None,
        *,
        rows=None,
        range=None,
        group_by=None,
        order_by=None,
    ) -> GroupedTable:
        """Apply a window over the input expressions.

        Parameters
        ----------
        window
            Window to add to the input
        rows
            Whether to use the `ROWS` window clause
        range
            Whether to use the `RANGE` window clause
        group_by
            Grouping key
        order_by
            Ordering key

        Returns
        -------
        GroupedTable
            A new grouped table expression
        """
        if window is None:
            window = ibis.window(
                rows=rows,
                range=range,
                group_by=group_by,
                order_by=order_by,
            )

        return self.__class__(
            self.table,
            self.by,
            having=self._having,
            order_by=self._order_by,
            window=window,
        )

    def count(self) -> ir.Table:
        """Computing the number of rows per group.

        Returns
        -------
        Table
            The aggregated table
        """
        table = self.table.to_expr()
        return table.aggregate(table.count(), by=self.groupings, having=self.havings)

    size = count


def _group_agg_dispatch(name):
    def wrapper(self, *args, **kwargs):
        f = getattr(self.arr, name)
        metric = f(*args, **kwargs)
        alias = f"{name}({self.arr.get_name()})"
        return self.parent.aggregate(metric.name(alias))

    wrapper.__name__ = name
    return wrapper


@public
class GroupedArray:
    def __init__(self, arr, parent):
        self.arr = arr
        self.parent = parent

    count = _group_agg_dispatch("count")
    size = count
    min = _group_agg_dispatch("min")
    max = _group_agg_dispatch("max")
    approx_nunique = _group_agg_dispatch("approx_nunique")
    approx_median = _group_agg_dispatch("approx_median")
    group_concat = _group_agg_dispatch("group_concat")


@public
class GroupedNumbers(GroupedArray):
    mean = _group_agg_dispatch("mean")
    sum = _group_agg_dispatch("sum")



from __future__ import annotations

import contextlib
import pathlib
from collections.abc import Callable  # noqa: TCH003
from typing import Annotated, Any, Optional

from public import public

import xorq.common.exceptions as com
from xorq.common.utils.env_utils import (
    EnvConfigable,
    env_templates_dir,
)
from xorq.loader import load_backend
from xorq.vendor.ibis.common.grounds import Annotable
from xorq.vendor.ibis.common.patterns import Between


PosInt = Annotated[int, Between(lower=0)]


class Config(Annotable):
    def get(self, key: str) -> Any:
        value = self
        for field in key.split("."):
            value = getattr(value, field)
        return value

    def set(self, key: str, value: Any) -> None:
        *prefix, key = key.split(".")
        conf = self
        for field in prefix:
            conf = getattr(conf, field)
        setattr(conf, key, value)

    @contextlib.contextmanager
    def _with_temporary(self, options):
        try:
            old = {}
            for key, value in options.items():
                old[key] = self.get(key)
                self.set(key, value)
            yield
        finally:
            for key, value in old.items():
                self.set(key, value)

    def __call__(self, options):
        return self._with_temporary(options)


class SQL(Config):
    """SQL-related options.

    Attributes
    ----------
    fuse_selects : bool
        Whether to fuse consecutive select queries into a single query where
        possible.
    default_limit : int | None
        Number of rows to be retrieved for a table expression without an
        explicit limit. [](`None`) means no limit.
    default_dialect : str
        Dialect to use for printing SQL when the backend cannot be determined.

    """

    fuse_selects: bool = True
    default_limit: Optional[PosInt] = None
    default_dialect: str = "duckdb"


class Interactive(Config):
    """Options controlling the interactive repr.

    Attributes
    ----------
    max_rows : int
        Maximum rows to pretty print.
    max_columns : int | None
        The maximum number of columns to pretty print. If 0 (the default), the
        number of columns will be inferred from output console size. Set to
        `None` for no limit.
    max_length : int
        Maximum length for pretty-printed arrays and maps.
    max_string : int
        Maximum length for pretty-printed strings.
    max_depth : int
        Maximum depth for nested data types.
    show_types : bool
        Show the inferred type of value expressions in the interactive repr.

    """

    max_rows: int = 10
    max_columns: Optional[int] = 0
    max_length: int = 2
    max_string: int = 80
    max_depth: int = 1
    show_types: bool = True


class Repr(Config):
    """Expression printing options.

    Attributes
    ----------
    depth : int
        The maximum number of expression nodes to print when repring.
    table_columns : int
        The number of columns to show in leaf table expressions.
    table_rows : int
        The number of rows to show for in memory tables.
    query_text_length : int
        The maximum number of characters to show in the `query` field repr of
        SQLQueryResult operations.
    show_types : bool
        Show the inferred type of value expressions in the repr.
    show_variables : bool
        Show the variables in the repr instead of generated names. This is
        an advanced option and may not work in all scenarios.
    interactive : bool
        Options controlling the interactive repr.
    """

    depth: Optional[PosInt] = None
    table_columns: Optional[PosInt] = None
    table_rows: PosInt = 10
    query_text_length: PosInt = 80
    show_types: bool = False
    show_variables: bool = False
    interactive: Interactive = Interactive()


env_config = EnvConfigable.subclass_from_env_file(
    env_templates_dir.joinpath(".env.xorq.template")
).from_env()


class Profiles(Config):
    profile_dir: pathlib.Path = pathlib.Path(env_config.XORQ_PROFILE_DIR).expanduser()


class Options(Config):
    """Ibis configuration options.

    Attributes
    ----------
    interactive : bool
        Show the first few rows of computing an expression when in a repl.
    repr : Repr
        Options controlling expression printing.
    verbose : bool
        Run in verbose mode if [](`True`)
    verbose_log: Callable[[str], None] | None
        A callable to use when logging.
    graphviz_repr : bool
        Render expressions as GraphViz PNGs when running in a Jupyter notebook.
    default_backend : Optional[ibis.backends.BaseBackend]
        The default backend to use for execution, defaults to DuckDB if not
        set.
    sql: SQL
        SQL-related options.
    clickhouse : Config | None
        Clickhouse specific options.
    dask : Config | None
        Dask specific options.
    impala : Config | None
        Impala specific options.
    pandas : Config | None
        Pandas specific options.
    pyspark : Config | None
        PySpark specific options.

    """

    interactive: bool = False
    repr: Repr = Repr()
    verbose: bool = False
    verbose_log: Optional[Callable] = None
    graphviz_repr: bool = False
    default_backend: Optional[Any] = None
    sql: SQL = SQL()
    clickhouse: Optional[Config] = None
    dask: Optional[Config] = None
    impala: Optional[Config] = None
    pandas: Optional[Config] = None
    pyspark: Optional[Config] = None
    profiles: Profiles = Profiles()


def _default_backend() -> Any:
    if (backend := options.default_backend) is not None:
        return backend

    try:
        import duckdb as _  # noqa: F401
    except ImportError:
        raise com.XorqError(
            """\
You have used a function that relies on the default backend, but the default
backend (DuckDB) is not installed.

You may specify an alternate backend to use, e.g.

ibis.set_backend("polars")

or to install the DuckDB backend, run:

    pip install 'ibis-framework[duckdb]'

or

    conda install -c conda-forge ibis-framework

For more information on available backends, visit https://ibis-project.org/install
"""
        )

    options.default_backend = con = load_backend("duckdb").connect(":memory:")
    return con


options = Options()


@public
def option_context(key, new_value):
    return options({key: new_value})


public(options=options)



from __future__ import annotations

import subprocess
from typing import Callable

import pandas as pd
import pytest
from pytest import param

import xorq.api as xo
import xorq.vendor.ibis.expr.types as ir
from xorq.tests.conftest import TEST_TABLES


def test_list_tables(con):
    tables = con.list_tables()
    assert isinstance(tables, list)
    key = "functional_alltypes"
    assert key in tables or key.upper() in tables
    assert all(isinstance(table, str) for table in tables)


def test_tables_accessor_mapping(con):
    if con.name == "snowflake":
        pytest.skip("snowflake sometimes counts more tables than are around")

    name = "functional_alltypes"

    assert isinstance(con.tables[name], ir.Table)

    with pytest.raises(KeyError, match="doesnt_exist"):
        con.tables["doesnt_exist"]

    # temporary might pop into existence in parallel test runs, in between the
    # first `list_tables` call and the second, so we check that there's a
    # non-empty intersection
    assert TEST_TABLES.keys() & set(map(str.lower, con.list_tables()))
    assert TEST_TABLES.keys() & set(map(str.lower, con.tables))


def test_tables_accessor_getattr(con):
    name = "functional_alltypes"
    assert isinstance(getattr(con.tables, name), ir.Table)

    with pytest.raises(AttributeError, match="doesnt_exist"):
        con.tables.doesnt_exist  # noqa: B018

    # Underscore/double-underscore attributes are never available, since many
    # python apis expect that checking for the absence of these to be cheap.
    with pytest.raises(AttributeError, match="_private_attr"):
        con.tables._private_attr  # noqa: B018


def test_tables_accessor_tab_completion(con):
    name = "functional_alltypes"
    attrs = dir(con.tables)
    assert name in attrs
    assert "keys" in attrs  # type methods also present

    keys = con.tables._ipython_key_completions_()
    assert name in keys


def test_tables_accessor_repr(con):
    name = "functional_alltypes"
    result = repr(con.tables)
    assert f"- {name}" in result


@pytest.mark.parametrize(
    "expr_fn",
    [
        param(lambda t: t.limit(5).limit(10), id="small_big"),
        param(lambda t: t.limit(10).limit(5), id="big_small"),
    ],
)
def test_limit_chain(alltypes, expr_fn):
    expr = expr_fn(alltypes)
    result = expr.execute()
    assert len(result) == 5


@pytest.mark.xfail
@pytest.mark.parametrize(
    "expr_fn",
    [
        param(lambda t: t, id="alltypes table"),
        param(lambda t: t.join(t.view(), [("id", "int_col")]), id="self join"),
    ],
)
def test_unbind(alltypes, expr_fn: Callable):
    xo.options.interactive = False

    expr = expr_fn(alltypes)
    assert expr.unbind() != expr
    assert expr.unbind().schema() == expr.schema()

    assert "Unbound" not in repr(expr)
    assert "Unbound" in repr(expr.unbind())


@pytest.mark.parametrize(
    ("extension", "method"),
    [("parquet", xo.read_parquet), ("csv", xo.read_csv)],
)
def test_read(data_dir, extension, method):
    table = method(
        data_dir / extension / f"batting.{extension}", table_name=f"batting-{extension}"
    )
    assert table.execute() is not None


@pytest.mark.parametrize(
    ("extension", "write", "read"),
    [
        (
            "parquet",
            xo.to_parquet,
            lambda path, schema: xo.read_parquet(
                path, schema=schema.to_pyarrow()
            ).execute(),
        ),
        (
            "csv",
            xo.to_csv,
            lambda path, schema: xo.read_csv(
                path, schema=schema.to_pyarrow()
            ).execute(),
        ),
        (
            "json",
            xo.to_json,
            lambda path, schema: pd.read_json(
                path, lines=True, dtype=dict(schema.to_pandas())
            ),
        ),
    ],
)
def test_write(alltypes, df, tmp_path, extension, write, read):
    output_path = tmp_path / f"alltypes.{extension}"
    write(alltypes, output_path)
    actual = read(output_path, alltypes.schema())

    assert output_path.exists()
    assert output_path.stat().st_size > 0
    assert isinstance(actual, pd.DataFrame)


@pytest.mark.parametrize(
    "method",
    [
        "deferred_read_csv",
        "deferred_read_parquet",
    ],
)
def test_deferred_read(method):
    assert hasattr(xo, method)


@pytest.mark.benchmark
@pytest.mark.parametrize(
    "module",
    ("xorq.api as xo",),
)
def test_benchmark_module_import(module):
    subprocess.check_output(
        (
            "python",
            "-c",
            f"import {module}",
        )
    )



import itertools
import logging
import time
from concurrent.futures import ThreadPoolExecutor
from queue import Queue

import pyarrow as pa
from cloudpickle import dumps, loads
from pyarrow.flight import (
    FlightClient as _FlightClient,
)

from xorq.common.utils.rbr_utils import (
    copy_rbr_batches,
)
from xorq.vendor import ibis


executor = ThreadPoolExecutor()


logger = logging.getLogger(__name__)


class FlightClient:
    def __init__(
        self,
        host="localhost",
        port=8815,
        username=None,
        password=None,
        cert_chain=None,
        private_key=None,
        tls_root_certs=None,
        healthcheck_n_tries=20,
        **kwargs,
    ):
        """
        Initialize the Ibis Backend Flight Client

        Args:
            host: Server host
            port: Server port
            username: User username
            password: User password
            tls_roots: TLS Root path
        """
        bytes_kwargs = {
            k: v
            for k, v in (
                ("cert_chain", cert_chain),
                ("private_key", private_key),
                ("tls_root_certs", tls_root_certs),
            )
            if v
        }
        scheme = "grpc+tls" if bytes_kwargs else "grpc"
        self._client = _FlightClient(
            f"{scheme}://{host}:{port}", **bytes_kwargs, **kwargs
        )
        self._wait_on_healthcheck(healthcheck_n_tries)

        if username and password:
            token_pair = self._client.authenticate_basic_token(
                username.encode(), password.encode()
            )
            self._options = pa.flight.FlightCallOptions(headers=[token_pair])
        else:
            self._options = None

    def _wait_on_healthcheck(self, n_tries=None, sleep_n=1):
        for attempt_i in itertools.islice(itertools.count(), n_tries):  # noqa: B007
            try:
                self.do_action(
                    "healthcheck",
                    options=pa.flight.FlightCallOptions(timeout=1),
                )
                logger.info("done healthcheck")
                return
            except pa.ArrowIOError as e:
                if "Deadline" in str(e):
                    logger.info("Server is not ready, waiting...")
                else:
                    raise e
            except pa.flight.FlightUnavailableError as e:
                matches = (
                    "Ssl handshake failed: SSL_ERROR_SSL: error:0A000086:SSL routines::certificate verify failed",
                    "Socket closed",
                )
                if any(to_match in str(e) for to_match in matches):
                    raise e
                else:
                    pass
            except pa.flight.FlightUnauthenticatedError:
                return
            except Exception:
                pass
            finally:
                logger.info(f"Flight server unavailable, sleeping {sleep_n} seconds")
                time.sleep(sleep_n)
        raise RuntimeError(f"failed to connect after {attempt_i} attempts")

    # FIXME: rename to execute_table, add execute that return pd.DataFrame
    def execute(self, expr, **kwargs):
        """
        Execute Expr and return results as Arrow table

        Args:
            expr: xorq Expression

        Returns:
            pa.Table
        """

        batches = self.execute_batches(expr, **kwargs)
        return batches.read_all()

    def execute_batches(self, expr, **kwargs):
        # Get FlightInfo
        flight_info = self._client.get_flight_info(
            pa.flight.FlightDescriptor.for_command(
                dumps(
                    {
                        "expr": expr,
                        **kwargs,
                    }
                )
            ),
            options=self._options,
        )

        # Get the first endpoint
        endpoint = flight_info.endpoints[0]

        # Get the result
        reader = self._client.do_get(endpoint.ticket, options=self._options)

        return reader

    def upload(self, table_name, expr, **kwargs):
        return self.upload_batches(
            self, table_name, expr.to_pyarrow_batches(), **kwargs
        )

    def upload_table(self, table_name, data, **kwargs):
        """
        Upload data to create or replace a table

        Args:
            table_name: Name of the table to create
            data: pa.Table containing the data
            **kwargs: Additional metadata (e.g., "target": "duckdb" or "iceberg")
        """
        command_payload = {"table_name": table_name, "kwargs": kwargs}

        writer, _ = self._client.do_put(
            pa.flight.FlightDescriptor.for_command(dumps(command_payload)),
            data.schema,
            options=self._options,
        )
        writer.write_table(data)
        writer.close()

    def upload_batches(self, table_name, reader, **kwargs):
        command_payload = {"table_name": table_name, "kwargs": kwargs}

        writer, _ = self._client.do_put(
            pa.flight.FlightDescriptor.for_command(dumps(command_payload)),
            reader.schema,
            options=self._options,
        )

        for batch in copy_rbr_batches(reader):
            writer.write_batch(batch)
        writer.done_writing()
        writer.close()

    def _do_action(self, action_type, action_body=None, options=None):
        if action_body is None:
            action_body = {}

        try:
            action = pa.flight.Action(
                action_type,
                dumps(action_body),
            )
            logger.info(f"Running action {action_type}")
            return map(
                loads,
                (
                    result.body.to_pybytes()
                    for result in self._client.do_action(
                        action, options=options or self._options
                    )
                ),
            )

        except pa.lib.ArrowIOError as e:
            logger.debug(f"Error calling action: {e}")

    def do_action_one(self, action_type, action_body=None, options=None):
        return next(
            self._do_action(
                action_type, action_body=action_body, options=options or self._options
            )
        )

    def do_action(self, action_type, action_body=None, options=None):
        return tuple(
            self._do_action(
                action_type, action_body=action_body, options=options or self._options
            )
        )

    def do_exchange_batches(self, command, reader):
        def do_writes(writer, reader):
            writer.begin(reader.schema)
            i = -1
            for i, batch in enumerate(reader, 1):  # noqa: B007
                writer.write_batch(batch)
            writer.done_writing()
            return i

        def do_reads(_reader, queue):
            i = -1
            for i, batch in enumerate(_reader, 1):  # noqa: B007
                queue.put(batch.data)
            queue.put(None)
            return i

        def do_writes_reads(command, reader, queue):
            descriptor = pa.flight.FlightDescriptor.for_command(command)
            writer, _reader = self._client.do_exchange(descriptor, self._options)
            # `with writer` must happen inside a future
            # # so its context remains alive during enclosed writes and reads
            with writer:
                do_writes_fut = executor.submit(do_writes, writer, reader)
                do_reads_fut = executor.submit(do_reads, _reader, queue)
                (n_writes, n_reads) = (do_writes_fut.result(), do_reads_fut.result())
            return {"n_writes": n_writes, "n_reads": n_reads}

        def queue_to_rbr(schema, queue):
            def queue_to_gen(queue):
                while (value := queue.get()) is not None:
                    yield value

            return pa.RecordBatchReader.from_batches(schema, queue_to_gen(queue))

        def get_output_schema(command, reader):
            (dct,) = self.do_action("query-exchange", command, options=self._options)
            assert dct["schema-in-condition"](ibis.schema(reader.schema))
            output_schema = dct["calc-schema-out"](
                ibis.schema(reader.schema)
            ).to_pyarrow()
            return output_schema

        queue = Queue()
        output_schema = get_output_schema(command, reader)
        fut = executor.submit(do_writes_reads, command, reader, queue)
        rbr = queue_to_rbr(output_schema, queue)
        return fut, rbr

    def do_exchange(self, command, expr):
        return self.do_exchange_batches(command, expr.to_pyarrow_batches())
